// REQUIRES: common.js



Webvisum.DB = {
    con : null,
    statements : null,
    init : function() {
        Webvisum.Common.log("initializing Webvisum DB");
        var file = Components.classes["@mozilla.org/file/directory_service;1"]
                .getService(Components.interfaces.nsIProperties).get("ProfD", Components.interfaces.nsIFile);
        file.append("webvisum.sqlite");
        try {
            var storageService = Components.classes["@mozilla.org/storage/service;1"]
                    .getService(Components.interfaces.mozIStorageService);
            this.con = storageService.openDatabase(file);
            if (!this.con.tableExists('objects')){
                // create table
                Webvisum.Common.log("Creating 'objects' table");
                //id INTEGER PRIMARY KEY AUTOINCREMENT
                this.con.createTable('objects', 
                    "id INTEGER PRIMARY KEY AUTOINCREMENT, domain VARCHAR(255) NOT NULL, type INTEGER NOT NULL, objectid VARCHAR(255) NOT NULL, lang VARCHAR(2) NOT NULL, label VARCHAR(255) NOT NULL");
                this.con.executeSimpleSQL("CREATE INDEX idx_objects ON objects (id ASC, domain ASC, type ASC, lang ASC)");
            }
            if (!this.con.tableExists('domains')){
                //create table
                Webvisum.Common.log("Creating 'domains' table");
                this.con.createTable('domains', 
                  'domain VARCHAR(255) PRIMARY KEY NOT NULL, updated_on INTEGER NOT NULL');
                this.con.executeSimpleSQL("CREATE UNIQUE INDEX idx_domains ON domains (domain ASC, updated_on DESC)");        
            }
            this.statements = {
                updateObject: this.con.createStatement("REPLACE INTO objects (domain, type, objectid, lang, label, id) VALUES (?1, ?2, ?3, ?4, ?5, ?6)"),
                updateDomain: this.con.createStatement("REPLACE INTO domains (domain, updated_on) VALUES (?1, ?2)"),
                selectObjectId: this.con.createStatement("SELECT id, label FROM objects WHERE domain = ?1 AND type = ?2 AND objectid = ?3 AND lang = ?4"),
                getObjects: this.con.createStatement("SELECT domain, lang, type, objectid, label, id  FROM objects WHERE domain = ?1 AND lang = ?2"),
                cleanObjects: this.con.createStatement("DELETE FROM objects WHERE domain = ?1 AND lang = ?2"),
                cleanDomain: this.con.createStatement("DELETE FROM domains WHERE domain = ?1"),
                deleteObject: this.con.createStatement("DELETE FROM objects WHERE id = ?1"),
                updateLabel: this.con.createStatement("UPDATE OR FAIL objects SET label=?4 WHERE domain = ?1 AND type = ?2 AND objectid = ?3 AND lang = ?4"),
                lastUpdate: this.con.createStatement("SELECT updated_on FROM domains WHERE domain = ?1")
            }
            return this
        }
        catch (e){
            Webvisum.Common.logErr("DB error:"+this.con.lastErrorString, e);
            return null;
        }
    }, 
    destroy : function() {
        Webvisum.Common.log("destroying Webvisum DB");
        this.con = null;
        for (var s in this.statements)
            this.statements[s].reset();
        this.statements = null;
    },
    dropAll : function() {
        this.con.executeSimpleSQL('DROP TABLE domains');
        this.con.executeSimpleSQL('DROP TABLE objects');
    },
    updateObject : function(aDomain, aType, anObjectId, aLabel, aLanguageCode) {
        const self = this;
        //composite PK not allowed by DB, should we make our own?
        const id = this.getObjectId(aDomain, aType, anObjectId, aLanguageCode);
        //Webvisum.Common.log('updating '+aDomain+" "+ anObjectId+ " "+aLanguageCode+ "  label:'"+aLabel+"' id="+id);
        if (aLabel && aLabel.length>0) {
            return this.ensure(this.statements.updateObject, function(statement) {
                statement.bindUTF8StringParameter(0, aDomain);
                statement.bindInt32Parameter(1, aType);
                statement.bindUTF8StringParameter(2, anObjectId);
                statement.bindUTF8StringParameter(3, aLanguageCode);
                statement.bindUTF8StringParameter(4, aLabel);
                if (id)
                    statement.bindInt32Parameter(5, id);
                statement.execute();
                return self.con.lastInsertRowID;
            })
        }
        else {
            return this.ensure(this.statements.deleteObject, function(statement) {
                statement.bindInt32Parameter(0, id);
                statement.execute();
            })
        }
    },
    cleanCache : function(aDomain, aLanguageCode) {
        this.ensure(this.statements.cleanDomain, function(statement) {
            statement.bindUTF8StringParameter(0, aDomain);
            statement.execute();
        })
        this.ensure(this.statements.cleanObjects, function(statement) {
            statement.bindUTF8StringParameter(0, aDomain);
            statement.bindUTF8StringParameter(1, aLanguageCode);
            statement.execute();
        })
    },
    updateLabel : function(aDomain, aType, anObjectId, aLabel, aLanguageCode) {
        return this.ensure(this.statements.updateLabel, function(statement) {
            statement.bindUTF8StringParameter(0, aDomain);
            statement.bindInt32Parameter(1, aType);
            statement.bindUTF8StringParameter(2, anObjectId);
            statement.bindUTF8StringParameter(3, aLabel);
            statement.bindUTF8StringParameter(3, aLanguageCode);
            statement.execute();
        })
    },
    getObjectId : function(aDomain, aType, anObjectId, aLanguageCode) {
        return this.ensure(this.statements.selectObjectId, function(statement) {
            statement.bindUTF8StringParameter(0, aDomain);
            statement.bindInt32Parameter(1, aType);
            statement.bindUTF8StringParameter(2, anObjectId);
            statement.bindUTF8StringParameter(3, aLanguageCode);

            if (statement.executeStep())
                return statement.getInt32(0);
            else
                return null
        })
    },
    updateDomain : function(aDomain, aTimestamp) {
        return this.ensure(this.statements.updateDomain, function(statement) {
            statement.bindUTF8StringParameter(0, aDomain);
            statement.bindInt64Parameter(1, aTimestamp);
            statement.execute();
        })
    },
    //returns int or null
    getLastUpdate : function(aDomain) {
        return this.ensure(this.statements.lastUpdate, function(statement) {
            statement.bindUTF8StringParameter(0, aDomain);
            if (statement.executeStep())
                return statement.getInt64(0);
            else
                return null
        })
    },
    //aLambda get an object with columns as properties
    getObjects : function(aDomain, aLambda) {

    },
    getAllObjects : function(aDomain, aLanguageCode) {
        return this.ensure(this.statements.getObjects, function(statement) {
            statement.bindUTF8StringParameter(0, aDomain);
            statement.bindUTF8StringParameter(1, aLanguageCode);
            const MAXRESULTS = 1000;
            var idx = 0;
            var results = [];
            while (statement.executeStep() && idx < MAXRESULTS) {
                results.push({
                    domain : statement.getUTF8String(0),
                    lang : statement.getUTF8String(1),
                    type : statement.getInt32(2),
                    objectid : statement.getUTF8String(3),
                    label : statement.getUTF8String(4),
                    id : statement.getInt32(5)
                });
                idx++;
            }
            return results;
        })
    },
    ensure : function(aStatement, aLambda) {
        try {
            return aLambda(aStatement);
        } catch (e) {
            Webvisum.Common.logErr("DB error:" + this.con.lastErrorString, e);
            return null;
        } finally {
            aStatement.reset();
        }

    },
    test : function() {
        this.updateObject("domena1", 3, "form#field", "aLabelismus");
        this.updateDomain("domena1", new Date());
    }
}