// REQUIRE: webvisum, common



Webvisum.Keys = {
    listener:null,
    init:function(aWebvisum) {
        this.webvisum = aWebvisum;
        this.webvisum.log("initializing Key shortcuts manager");
        var e = document.getElementById("appcontent");
        if (e) {
            this.listener = Webvisum.Common.objMethod(this, 'onKeyPressed');
            addEventListener("keypress", this.listener, false);
        }
        return this;
    }, 
    destroy:function() {
        this.webvisum.log("removing Key shortcuts manager");
        var e = document.getElementById("appcontent");
        if(e && this.listener) { 
            removeEventListener("keypress", this.listener, false);
            this.listener = null;
        }
    }, 
    onKeyPressed:function (anEvent) {
      // log("KEY: char:" + anEvent.charCode + "  code:" + anEvent.keyCode);
      const bindings = this.keyBindings;
      var cmd, p, sk, mods;
      for (var e in bindings) {
        cmd = e;
        sk = bindings[e];
        if (anEvent.charCode && anEvent.charCode == sk.charCode || anEvent.keyCode && anEvent.keyCode == sk.keyCode) {
          mods = sk.modifiers;
          for (p in mods) {
            if (anEvent[p] != mods[p]) {
              cmd = null;
              break;
            }
          }          
          if (cmd) {
            anEvent.preventDefault();
            this.executeCommand(cmd);
            return;
          }
        }
      }
    },
    executeCommand:function (aCommand) {
        Webvisum.Common.log('executing key shortcut action:' +aCommand);
        this.webvisum.executeCommand(aCommand)
    },
    keyBindings : {
        help: null,
        menu: null,
        label: null,
        labelpage: null,
        ocr: null,
        captcha: null,
        lastlink: null,
        highlight: null,
        underline: null,
        hideimages: null,
        visuals: null,
        sounds: null,
        pageinfo: null,
        pageupdate: null,
        status: null
        //activity: null
    },
    Key : function(){
        this.modifiers = { 
            shiftKey: false, 
            altKey: false, 
            metaKey: false, 
            ctrlKey: false 
        };
        this.charCode = 0; 
        this.keyCode = 0;
        this.toString = function() {
            return "Key: modifs"+Webvisum.Common.dumpObj(this.modifiers)+" keyCode:"+this.keyCode+" keyChar:"+this.charCode;
        }
    },
    isValid : function(anEventName) {
        return anEventName in this.keyBindings
    },
    keyFromString : function(aKeyDefinition) {
        var values = Webvisum.Common.trim(aKeyDefinition).split(/\s+/);
        var vpos = values.length;
        if (vpos) {            
            var key = new Webvisum.Keys.Key();
            // modifiers
            var keyVal = values[--vpos];
            for (var value; vpos-- > 0;) {
                var value = values[vpos] + "Key";
                if (value in key.modifiers) key.modifiers[value] = true;
            }
            // key code
            if (keyVal.length > 1) {
                key.keyCode = KeyEvent["DOM_VK_" + keyVal.toUpperCase()] || 0;
            } else {
                key.charCode = (key.modifiers.shiftKey ? keyVal.toUpperCase() : keyVal).charCodeAt(0) || 0;
            }
            return key;
        } else {
            return null;
        }
    },
    setKeyBinding:function (aCommand, aKeyDefinition) {
        const key = this.keyFromString(aKeyDefinition);
        Webvisum.Common.log('Setting key binding for event '+aCommand +' to "'+aKeyDefinition+'" ='+key);
        if (key) {            
            this.keyBindings[aCommand] = key;
        } else {
            delete(this.keyBindings[aCommand]);
        }
    }    
}
