// REQUIRES: common.js



/*
 * Abstract Webvisum Object - represents an object on a page that can be labeled
 */
Webvisum.Object = {
    type:null,
    id:null,
    label:null,
    page_url:null,
    language:null,
    isDirtyDocument: false,
    isDirtyCache: false,
    isDirtyServer: false,
    
    name:null,   //transient 
    targets:null, //transient
    get target(){
      if (this.targets && this.targets.length>0) return this.targets[0];
      return null;
    },
    get targetCount(){
      return this.targets ? this.targets.length : 0;   
    },
    forceTarget:function (aTarget) {
        this.targets =  aTarget ? [aTarget] : null
    },
    fromArgs:function (aType, anId, aLabel, aLanguage, aPageUrl) {
        this.type = aType; 
        this.id = anId;        
        this.label = aLabel;
        this.page_url = aPageUrl;
        this.language = aLanguage;
        //this.target = this.resolveTarget(); - neeed to ce called explicitely
    },
    fromXML:function (anElement, aPageUrl, aLanguage) {
        this.type = anElement.getAttribute('type');        
        this.id = anElement.getAttribute('id');        
        this.label = anElement.getAttribute('label');
        this.page_url = aPageUrl;
        this.language = aLanguage;
        //this.target = this.resolveTarget(); - neeed to ce called explicitely
    },
    fromDOM:function (anTargetElement, aContext) {
        //subclass responsibility
    },
    verifyTarget:function(aWebvisumDocument) {
        //resolve targets if necessary
        if (!this.target) {            
            var trg = this.resolveTarget(aWebvisumDocument);
            if (trg instanceof Array) this.targets = trg;
            else this.targets =  trg ? [trg] : null;
            Webvisum.Common.log('Found '+this.targetCount+' '+this.name+' targets');
        }
        if (!this.target) {
            Webvisum.Common.log('[NO TARGET] target found for '+this.name+' object:', this.toString()+"   =>  "+aWebvisumDocument) ;
            return false;
        }
        return true;        
    },
    apply:function (aWebvisumDocument) { 
        if (!aWebvisumDocument) aWebvisumDocument = this.binding; //sometimes there is no context given, because it was set previously (eg. last anchors)
        if (!this.verifyTarget(aWebvisumDocument)) return;
        for (var targetIdx = 0; targetIdx < this.targets.length; targetIdx++) {
            try{
                this.applyTarget(this.targets[targetIdx], aWebvisumDocument);
            }
            catch (e) {
                Webvisum.Common.logErr("Could not apply object '"+this+"' to target '"+this.targets[targetIdx]+"'");
            }
        }
    },
    applyTarget:function (aTarget, aWebvisumDocument) {
        //subclass responsibility 
    },
    resolveTarget:function (aWebvisumDocument) { 
        this.binding = aWebvisumDocument;
        this.binding.addObject(this);
        //subclass responsibility
    },
    editLabel:function (aLabel) {
        if (aLabel!=this.label){
            this.label = aLabel;
            this.isDirtyDocument = true;
            this.isDirtyCache = true;
            this.isDirtyServer = true;
            return true;
        }
        return false;
    },
    Factory:new Array(),
    Enlist:function (aWebvisumObject) {
        this.Factory[aWebvisumObject.type]=aWebvisumObject;
        this.Factory[aWebvisumObject.name]=aWebvisumObject;
        return this;
    },
    FactorByTypeOrName:function (aTypeOrName) {
        const proto = this.Factory[aTypeOrName];
        if (proto) return Webvisum.beget(proto);
        return null;
    },
    FactorFromXML:function (anElement, aPageUrl, aLanguage) {
        const type = anElement.getAttribute('type');        
        var newObject = this.FactorByTypeOrName(type); 
        newObject.fromXML(anElement, aPageUrl, aLanguage);    
        return newObject;
    },
    FactorFromArgs:function (aType, anId, aLabel, aLanguage, aPageUrl) {
        var newObject = this.FactorByTypeOrName(aType); 
        newObject.fromArgs(aType, anId, aLabel, aLanguage, aPageUrl);    
        return newObject;
    },
    _dumpTargets:function () { 
        var tgs = this.targets;
        if (!tgs) return 'N/A';
        var msg = '[';
        for (var i = 0; i < tgs.length; i++) {
            if (i>0) msg+=',';
            msg+=tgs[i].nodeName;
        }
        msg+=']';
        return msg
        
    },
    toString:function() {
        return this.name+' ['+this.id+"] @ "+this.page_url+" ("+this.language+")='"+this.label
            +"' dirty=["+this.isDirtyDocument+","+this.isDirtyCache+","+this.isDirtyServer+"]" + " targets:"+this._dumpTargets();
    },
    Utils:{
        generateId:function (anObject) {
            return 'ws_'+anObject.type+'_'+anObject.id;
        },

        generateLabel:function (anObject, parentNode) {
            Webvisum.Monkey.html({
                'span': {
                    'class': 'webvisum',
                    'id':  this.generateId(anObject),
                    '#text': anObject.label
                }
            }, parentNode);
        },

        // http://fero.com/xx, http://test.com/file -> http://fero.com/xx
        // /fero/xx, http://test.com/file -> /fero/xx
        // http://test.com/xx, http://test.com/file -> /xx
        // http://www.test.com/xx, http://test.com/file -> /xx
        normalizeUri:function (aUriString, aDocument) {
            const docUri = Webvisum.Monkey.createUri( aDocument.location );
            const uri = Webvisum.Monkey.createUri(aUriString, docUri);
            const docDomain = Webvisum.Monkey.extractDomainFromUrl(docUri);
            const uriDomain = Webvisum.Monkey.extractDomainFromUrl(uri);
//            Webvisum.Common.log("uri="+aUriString+"   aDocument"+aDocument);
//            Webvisum.Common.log( docUri +" doc\n"
//                +uri+" uri\n"
//                +docDomain+" docdom\n"
//                +uriDomain+ "uriDom");
            if (docDomain == uriDomain){
                return uri.path
            }
            else return aUriString
        },
        generateLabelDOM:function (anObject, aTarget) {
            const doc = aTarget.ownerDocument;
            const insert = doc.createElement('span');
            insert.setAttribute('class', 'webvisum');
            insert.setAttribute('id', this.generateId(anObject));
            insert.textContent = anObject.label;
            return insert;
        },
        getXPathElements:function (anXPath, aContext) {
            Webvisum.Common.log('evaluating xpath:'+anXPath+ 'on context '+aContext);
            const doc = aContext.ownerDocument || aContext;
            return doc.evaluate(anXPath, aContext, null, XPathResult.UNORDERED_NODE_ITERATOR_TYPE, null );
        },
        fromXPathId:function (anObjectId) {
            var match = anObjectId.match(/^xpath\:\/\/(.*)$/);
            return match ? match[1] : null; 
        },
        toXPathId:function (anXPath) {
            return "xpath://"+anXPath
        },
        existsInParent:function (anObject, aParanetElement) {
            return Webvisum.Common.idd( this.generateId(anObject), aParanetElement );
        }
    }
}

Webvisum.ObjectLink = Webvisum.beget(Webvisum.Object);
Webvisum.ObjectLink.include({
    type:1,
    name:'link',
    resolveTarget:function (aWebvisumDocument) {
        Webvisum.Object.resolveTarget.call(this, aWebvisumDocument); //static super
        
        //this.superMethod('resolveTarget', aWebvisumDocument);
        const doc = aWebvisumDocument.document;
        const links = doc.links;
        var matches = [];
        Webvisum.Common.log("Link, resolving from "+links.length+" links");
        for (var i = 0; i < links.length; i++) {
            var hrefUri = this.generateId( links[i].href, doc);
            // Webvisum.Common.log(hrefUri + "  ==== "+ this.id);
            if (hrefUri == this.id) {
                matches.push( links[i] );
            }
        }        
        return matches;
    },    
    fromDOM:function (anAnchorElement, aContext) {
        const srctarget = anAnchorElement;
        const wdoc = aContext.wdocument;
        this.page_url = aContext.pageURL;
        const doc = wdoc.document;
        this.id = this.generateId( srctarget.href.toString(), doc );        
        this.label = srctarget.textContent;
        this.language = aContext.language;
        return this.id ? this : null;
    },
    applyTarget:function (aTarget, aWebvisumDocument) {
        Webvisum.Object.Utils.generateLabel(this, aTarget);
        aTarget.setAttribute('title', this.label);
        this.isDirtyDocument = false;
    },
    getLocalizedName:function() {
        return 'Link';
    },
    generateId:function (aHrefUrl, aDocument) {
        if (!aHrefUrl || aHrefUrl.length==0 || aHrefUrl=='#'){
            //self reference
            return null; //TODO:  how this should be handled?
        }
        else if (/^javascript:/.test(aHrefUrl)){
            //javascript href
            return "js:"+Webvisum.Monkey.hashString(aHrefUrl);
        }
        else {            
            return Webvisum.Object.Utils.normalizeUri( aHrefUrl, aDocument);
        }
    }
});
Webvisum.Object.Enlist(Webvisum.ObjectLink);

Webvisum.ObjectImage = Webvisum.beget(Webvisum.Object);
Webvisum.ObjectImage.include({
    type:2,
    name:'image',
    resolveTarget:function (aWebvisumDocument) {
        Webvisum.Object.resolveTarget.call(this, aWebvisumDocument); //static super
        const doc = aWebvisumDocument.document;
        const imgs = doc.images;
        var matches = [];
        Webvisum.Common.log("Link, resolving from "+imgs.length+" images");
        for (var i = 0; i < imgs.length; i++) {
            var imgUrl = Webvisum.Object.Utils.normalizeUri(imgs[i].src, doc);
            // Webvisum.Common.log(imgUrl + "  ==== "+ this.id);
            if (imgUrl==this.id) matches.push(  imgs[i] );
        }
        return matches;
    },        
    fromDOM:function (anImage, aContext) {
        const srctarget = anImage;
        const wdoc = aContext.wdocument;
        this.page_url = aContext.pageURL;
        const doc = wdoc.document;
        this.id = Webvisum.Object.Utils.normalizeUri( srctarget.src, doc);
        this.label = srctarget.title || srctarget.alt;
        this.language = aContext.language;
        return this;
    },
    applyTarget:function (aTarget, aWebvisumDocument) {
        const parent = aTarget.parentNode;        
        const insert = Webvisum.Object.Utils.generateLabelDOM(this, aTarget);
        const existing = Webvisum.Object.Utils.existsInParent(this, parent);        
        
        if (existing) parent.replaceChild(insert, existing);
        else parent.insertBefore(insert, aTarget);
        aTarget.setAttribute('alt', this.label );
        aTarget.style.display='none';
        this.isDirtyDocument = false;
    },
    getLocalizedName:function() {
        return 'Image';
    }
}); 
Webvisum.Object.Enlist(Webvisum.ObjectImage);

Webvisum.ObjectForm = Webvisum.beget(Webvisum.Object);
Webvisum.ObjectForm.include({
    type:3,
    name:'form',    
    resolveTarget:function (aWebvisumDocument) {
        Webvisum.Object.resolveTarget.call(this, aWebvisumDocument); //static super
        const doc = aWebvisumDocument.document;
        const xpath = Webvisum.Object.Utils.fromXPathId(this.id);
        if (xpath){
            var iterator = Webvisum.Object.Utils.getXPathElements(xpath, doc);
            var matches = [];
            try {
                var node = iterator.iterateNext();
                while (node) {
                    matches.push(node);
                    node = iterator.iterateNext();
                } 
            }
            catch (e) {} //swallow doc modification
            return matches;
        }
        else {
            //old method
            const parsedId = this.id.split('#');
            const frm = Webvisum.Common.unwrap(doc.forms.namedItem(parsedId[0]));
            return frm ? frm[parsedId[1]] : null;            
        }
    },        
    fromDOM:function (aFormField, aContext) {
        const srctarget = aFormField;
        this.page_url = srctarget.ownerDocument.location.toString();
        this.id = this.generateId( aFormField );
//        this.id = srctarget.form.getAttribute('name')+"#"+ srctarget.name;
        this.label = null;
        this.language = aContext.language;
        return this;
    },
    applyTarget:function (aTarget, aWebvisumDocument) {
        const parent = aTarget.parentNode;
        const insert = Webvisum.Object.Utils.generateLabelDOM(this, aTarget);
        const existing = Webvisum.Object.Utils.existsInParent(this, parent);
        Webvisum.Common.log(">>>>>"+existing+"  "+this);
        if (existing) parent.replaceChild(insert, existing);
        else parent.insertBefore(insert, aTarget);        
        aTarget.setAttribute('title', this.label );
        this.isDirtyDocument = false;
    },
    getLocalizedName:function() {
        return 'Form Field';
    },
    generateId:function (anInputElement) {
        const targetTag = anInputElement.nodeName.toLowerCase();
        const formAction = anInputElement.form.getAttribute('action'); 
        const formName = anInputElement.form.getAttribute('name');
        const targetName = anInputElement.getAttribute('name');
        
        var c=[];
        if (formAction) c.push( '@action="'+formAction+'"' );
        if (formName) c.push( '@name="'+formName+'"' );
        
        const formCondition = c.length>0 ? '['+c.join(' and ')+']' : '';        
        const targetCondition = targetName ? ('[@name="'+targetName+'"]') : '';                 
        const xpath = '//form'+formCondition+'//'+targetTag+targetCondition;        
        Webvisum.Common.log('generated: '+xpath);
        
        //test
        var iterator = Webvisum.Object.Utils.getXPathElements(xpath, anInputElement.ownerDocument);
        var node = iterator.iterateNext();
        while (node) {
            Webvisum.Common.log("TEST:"+node.name);
            node = iterator.iterateNext();
        } 
        
        return Webvisum.Object.Utils.toXPathId(xpath);
        
    }
});
Webvisum.Object.Enlist(Webvisum.ObjectForm);

Webvisum.ObjectPage = Webvisum.beget(Webvisum.Object);
Webvisum.ObjectPage.include({
    type:4,
    name:'page',
    resolveTarget:function (aWebvisumDocument) {
        Webvisum.Object.resolveTarget.call(this, aWebvisumDocument); //static super
        var doc = aWebvisumDocument.document;
        if (doc.location.toString()==this.page_url)
            return  doc;
        else 
            return null;
    },        
    fromDOM:function (aDocument, aContext) {
        const srctarget = aDocument;
//        this.binding = aContext.wdocument;
        this.page_url = srctarget.location.toString();
        this.id = this.page_url;
        this.label = srctarget.title;
        this.language = aContext.language;
        return this;
    },
    applyTarget:function (aTarget, aWebvisumDocument) {
        aTarget.title = this.label;
        this.isDirtyDocument = false;
    },
    getLocalizedName:function() {
        return 'Page Title';
    }
});
Webvisum.Object.Enlist(Webvisum.ObjectPage);


Webvisum.CaptchaRules = {
    xmlDocument:null,
    rules:[],
    last_update:0,
    min_weight:0,
    fromXml:function (aDocument) {
        this.xmlDocument = aDocument;
        //parse xml
        this.last_update = 1*Webvisum.Common.tag('last_update',this.xmlDocument).textContent;
        this.min_weight = 1*Webvisum.Common.tag('min_weight',this.xmlDocument).textContent;
        const elems = Webvisum.Common.tags('rule', this.xmlDocument);
        this.rules = [];
        if (elems.length==0) return null;
        for (var i = 0; i < elems.length; i++) {
            var rule = {
                pattern: elems[i].getAttribute('regexp'),
                weight: 1*elems[i].getAttribute('points')
            }
            if (rule.pattern.length>0) {
                this.rules.push(rule);
            }
        }
        return this;
    },
    fromString:function (aXmlString) {
        var parser = new DOMParser();
        var dom = parser.parseFromString(aXmlString, "text/xml");
        this.fromXml(dom);
        return this;
    },
    toXmlString:function () {
        return this.xmlDocument ? new XMLSerializer().serializeToString(this.xmlDocument) : null;
    },
    toString:function () {
        return "last_update:" + this.last_update+ "="+Webvisum.Common.toDate(this.last_update)+", min_weight:"+
            this.min_weight+", rules:"+this.rules.length+" "+Webvisum.Common.dumpObj(this.rules);   
    }
}


Webvisum.Document = function(aLocalWebvisum, aDocument){
    this.webvisum = aLocalWebvisum;
    this.document = Webvisum.Common.unwrap(aDocument);
    this.isHandled = false;
    this.isBlocked = false;
    this._domain = null;
    this.meta = {annotationSrc:'none'}
    this.objects = {} //id->aWebvisumObject
    this.init();
}

Webvisum.Document.prototype = {
    init:function () {
        this.page_url = this.document.location;
        this.window = Webvisum.Common.unwrap(this.document.defaultView);
    },
    get domain(){
        if (!this._domain){
            this._domain = Webvisum.Monkey.extractDomainFromUrl(this.page_url); 
        }
        return this._domain;
    },
    addObject:function (aWebvisumObject) {
        this.objects[aWebvisumObject.id]  = aWebvisumObject;
    },
    toString:function () {
        return "WDocument "+this.page_url;
    }
}
