// REQUIRES: urlfilter.js, common.js, sound.js

Webvisum.Preferences = {

  SITE_BASE_URL:  'https://www.webvisum.com/',
  REGISTER_URL: 'https://www.webvisum.com/main/register/popup',
  FEEDBACK_URL: 'https://www.webvisum.com/main/feedback/popup',
  FORGOTPASS_URL: 'https://www.webvisum.com/main/forgot/popup',

  selectFile: function(aTextFieldId) {
      try {
          const FILEPICKER = Components.interfaces.nsIFilePicker;
          const picker = Components.classes["@mozilla.org/filepicker;1"]
                  .createInstance(FILEPICKER);
          picker.init(window, null, FILEPICKER.modeOpen);
          picker.appendFilters(FILEPICKER.filterAll);

          if (picker.show() == FILEPICKER.returnOK) {
              var field = Webvisum.Common.idd(aTextFieldId);
              field.value = 'file://'+picker.file.path;
          }
      } catch (e) {
          Webvisum.Common.logErr("Cannot select file", e);
      }
  },

  verifyCorectSoundPath: function(aPath){
      if (Webvisum.getGlobalWebvisum().sound._toUrl(aPath)==null) {
          var msg = Webvisum.Preferences.m('msg.webvisum.sound.path_error', [aPath]);        
          alert(msg);
      }
  },

  testSound: function(aTextboxId) {
      Webvisum.getGlobalWebvisum().sound.playStringUrl(Webvisum.Common.idd(aTextboxId).value)
  },

  defaultSounds: function() {
      try {
          Webvisum.Common.idd('pref_sound_click').reset();
      } catch (e) {
      };
      try {
          Webvisum.Common.idd('pref_sound_error').reset();
      } catch (e) {
      };
  },

  testTriggerUrl: function() {
      const testUrl = Webvisum.Common.idd('trigger.test.url').value;
      var filter = Webvisum.beget(Webvisum.UrlFilter);
      const whitelist = Webvisum.Common.idd('in_whitelist').value;
      const blacklist = Webvisum.Common.idd('in_blacklist').value;
      filter.setRules(whitelist, blacklist);
      var msg = (filter.isAllowed(testUrl)) ? 
          Webvisum.Preferences.m('label.webvisum.trigger.allowed') : 
          Webvisum.Preferences.m('label.webvisum.trigger.blocked');
      Webvisum.Common.alertEx(msg, 'Webvisum' );
  },

  defaultTrigger: function() {
      try {
          Webvisum.Common.idd('pref_blacklist').reset();
      } catch (e) {
      };
      try {
          Webvisum.Common.idd('pref_whitelist').reset();
      } catch (e) {
      };
  },

  openRegistation: function() {
      window.openDialog(Webvisum.Preferences.REGISTER_URL, "webvisum_registration",
          "titlebar,centerscreen,resizable");
  },

  openForgotPwd: function() {
      window.openDialog(Webvisum.Preferences.FORGOTPASS_URL, "webvisum_forgotpwd",
              "titlebar,centerscreen,resizable");
  },

  openFeedback: function() {
      window.openDialog(Webvisum.Preferences.FEEDBACK_URL, "webvisum_feedback",
              "titlebar,centerscreen,resizable");
  },

  EventKey: {},

  KeyCapture: {
      listener : null,
      textbox : null,
      oldvalue: null,

      toggle : function(anObject) {
          if (this.listener)
              this.stop(anObject);
          else
              this.start(anObject);
      },
      start : function(anObject) {
          Webvisum.Common.log('Starting key capture');
          this.listener = Webvisum.Common.objMethod(this, 'inputKeydown');
          this.textbox = Webvisum.Common.idd('in_binding');
          this.oldvalue = this.textbox.value; 
          this.textbox.value = '';
          anObject.addEventListener("keypress", this.listener, false);
          anObject.label = Webvisum.Preferences.m('label.webvisum.key.stopcapture');
          anObject.setAttribute('class', 'recording');
      },
      stop : function(anObject) {
          Webvisum.Common.log('Stopping key capture');
          anObject.removeEventListener("keypress", this.listener, false);
          anObject.label = Webvisum.Preferences.m('label.webvisum.key.capture');
          anObject.setAttribute('class', '');
          this.textbox = Webvisum.Common.idd('in_binding');
          if (this.textbox.value=='') this.textbox.value=this.oldvalue;
          this.listener = null;
      },
      inputKeydown : function(anEvent) {
          //Webvisum.Common.printObj(anEvent);  
          var modif = [];
          if (anEvent.altKey)
              modif.push('alt');
          if (anEvent.metaKey)
              modif.push('meta');
          if (anEvent.shiftKey)
              modif.push('shift');
          if (anEvent.ctrlKey)
              modif.push('ctrl');
          var modifString = modif.join(' ');
          var key = anEvent.charCode > 0
                  ? String.fromCharCode(anEvent.charCode)
                  : null;
          if (!key) {
              key = Webvisum.Preferences.EventKey[anEvent.keyCode];
          }

          Webvisum.Common.log("KEY: char:" + anEvent.charCode + "  key:" + anEvent.keyCode
                  + "  isChar" + anEvent.isChar + "  modif:" + modifString);
          //        Webvisum.Common.printObj(anEvent);
          this.textbox.value = modifString.length > 0
                  ? modifString + " " + key : key;
          this.storeBinding(this.textbox.value);                
          this.textbox.focus();
          anEvent.stopPropagation();
          anEvent.preventDefault();
          this.stop(anEvent.target);
          return false;
      },
      keySelect : function(aMenuItem) {
          var event = aMenuItem.value;
          var textbox = Webvisum.Common.idd('in_binding');
          var value = Webvisum.getGlobalWebvisum().prefs.getCharPref('keys.' + event);
          Webvisum.Common.log("selected event:" + event + "  value=" + value);
          textbox.value = value;
      },
      storeBinding : function(aValue) {
          const event = Webvisum.Common.idd('pref_key_event').selectedItem.value;
          Webvisum.Common.log('storing keybinding '+event+'->'+aValue);
          Webvisum.getGlobalWebvisum().prefs.setCharPref('keys.' + event, aValue);
      }
  },

  defaultKeys: function() {
      var prefs = Webvisum.getGlobalWebvisum().prefs;
      for(var event in Webvisum.Keys.keyBindings) {
          if (prefs.prefHasUserValue('keys.' + event))
              prefs.clearUserPref('keys.' + event);
      }
      // refresh
      Webvisum.Preferences.KeyCapture.keySelect(Webvisum.Common.idd('pref_key_event').selectedItem);
  },

  onInit: function() {
      for (k in KeyEvent) {
          Webvisum.Preferences.EventKey[KeyEvent[k]] = k.substring(7); // remove DOM_VK_
      }
      sizeToContent();
      Webvisum.Common.LOGGING = Webvisum.getGlobalWebvisum().isDebug(); // global hack for logging
      
      Webvisum.Preferences.KeyCapture.keySelect(Webvisum.Common.idd('pref_key_event').selectedItem);    
  },

  m: function(aString, optArgArray) {
      var stringsBundle = document.getElementById("webvisum-pref-strings");
      return optArgArray ?
          stringsBundle.getFormattedString(aString, optArgArray):
          stringsBundle.getString(aString);
  },

  onAccept: function(){
    //if set, save the new password 
    var pwdEle = document.getElementById('pwd');
    var pwd = Webvisum.Common.trim(pwdEle.value);
    if (pwd && !Webvisum.Monkey.savePassword(pwd)){
      //user refused to enter their master password, so our password was unable to be saved into the Login Manager
      alert(Webvisum.Preferences.m("msg.webvisum.savepassworderror"));
      
      //focus on the password box
      document.getElementById('webvisum-preferences').showPane(document.getElementById('pp_general'));
      pwdEle.select();
      
      //and leave the dialog open (user can then choose to clear the password and save any other preferences, or enter the master password)
      return false;
    }
  }
}
