// REQUIRES: common.js (state.js )



Webvisum.Server = function(aStateManager, aLogProvider) {
    this.state = aStateManager;
    this.logger = aLogProvider;
    this.disabled = true;
    this.genericRequest = Webvisum.Request; //Webvisum.beget(Webvisum.Request);
    //key->QueueEntry
    this.queue = {};
    this.QueueEntry = function(aKey, aRequest, aContext, anOpts) {
        this.opts = {
            submissive: false, //if any request with the same key is already in queue, a new submissive one will be knockedout 
            delay: 0,  //execution delay for request buffering, 0 = immediate execution; [Integer] in milliseconds
            frequencyGuard: -1  // a timeout [ms], for how long the same request will not be executed
        }
        Webvisum.Common.merge(this.opts, anOpts);
        this.key = aKey;
        this.request = aRequest;
        this.context = aContext;
        this.state = null;
    };
    this.QueueEntry.prototype = {
        toString : function() {
            return "'" + this.key + "'=" + this.request + " [" + this.state
                    + "] opts:"+Webvisum.Common.dumpObj(this.opts);
        }
    }
    
    //request key -> timeout time
    this.frequencyGuard = {}
}

Webvisum.Server.prototype = {
    /* 
     * Set basic server settings 
     */
    setServer : function(aHost, aPort, aUsername, aPwd) {
        this.genericRequest.init(aHost, aPort, aUsername, aPwd);
    },
    // communucatuion is operative
    enable:function () {
        this.logger.log('Enabling server');
        this.disabled = false;
    },
    // no communication AT ALL will happen 
    disable:function () {
        this.logger.log('Disabling server');
        this.disabled = true;
    },
    /* create new request from the generic request
     * Optional Method [String]
     * optional Params [Object<string,string>]
     * optional prototyoe aProto [descendant of Webvisum.Request] 
     */
    newRequest : function(aMethod, aParams, aProto /*optional*/) {
        const proto = aProto ? aProto : this.genericRequest;
        var rq = Webvisum.beget(proto);
        rq.method = aMethod;
        if (aParams)
            rq.params = aParams;
        return rq;
    },
    /* 
     * Send request
     * aReq [Webvisum.Request] request
     * aContext - any information/object to be passed to the response
     * */
    sendRequest : function(aRequest, aContext) {
        var url = aRequest.toUrl();
        return this._send(aRequest.http_method, url, aRequest.data,
                aRequest.headers, aRequest.responseSuccessFn,
                aRequest.responseFailureFn, aContext);
    },
    /*
     * Queue Request - place the request to an execution buffer (not really a queue)
     * Used to protect server form unnecessary load.
     * opts (optional) = {submissive:false, delay: 0} 
     */
    queueRequest : function(aRequest, aContext, anOpts) {
        var queue = this.queue;
        const key = aRequest.key;
        
        const entry = new this.QueueEntry(key, aRequest, aContext, anOpts);
        
        // frequency Guard check
        if (!this.passesFrequencyGuard(key, entry.opts.frequencyGuard)){
            this.logger.log('Rejecting query due to frequeccyGuard: '+entry);
            return
        }
        
        const delay = entry.opts.delay;
        if (delay == 0) {
            this.sendRequest(aRequest, aContext);
            return;
        }
        entry.request.frozen = true; //dissallow key/delay modification
        this.logger.log("Queuing reqest. " + entry);
        //if found replace it and set new expiration; mark old as obsolete
        if (key in queue) {
            if (entry.opts.submissive) {
                this.logger.log('Discarding submissive entry: '+entry);
                return;
            }
            else {
                this.logger.log('Replacing old request in the queue');
                queue[key].state = 'obsolete';
            }
        }
        queue[key] = entry;
        const server = this;
        const self = this;
        const trigger = (function() {
            self.logger.log('Trigger of ' + entry)
            if (entry.state != 'obsolete') {
                entry.state = 'running';
                delete queue[key];
                server.sendRequest(entry.request, entry.context);
            }
        });
        setTimeout(trigger, delay);
    },
    
    //aTimeout = -1  -> nothing is checked nor set
    //aTimeout = 0  -> it is checked but not set
    //aTimeout >0  -> checked and set
    passesFrequencyGuard:function (aKey, aTimeoutMs) {
        if (aTimeoutMs!=null && aTimeoutMs<0) return true; 
        const currentTime = 1*new Date();
        
        var passed = true;
        // chech key  & garbage collect the guard
        for (var key in this.frequencyGuard){
            var timeout = this.frequencyGuard[aKey];
            if (currentTime>timeout) {
                delete this.frequencyGuard[aKey];
            }            
            else if (key==aKey) passed = false;
        }
        //set new timeout
        if (passed && aTimeoutMs>0)
            this.frequencyGuard[aKey]= currentTime+aTimeoutMs;
            
        return passed;
    },
    _send : function(aMethod, aUrl, aData, aHeaders, aSuccessHandlerFn,
            aFailureHandlerFn, aContext) {
        if (this.disabled) {
            this.logger.log('Server is DISABLED, '+aMethod+" "+aUrl);
            return;
        }
                
        this.state.setState('connect', true);
        var req = new XMLHttpRequest();
        req.overrideMimeType('text/xml');
        req.onload = this._generateHandler(aSuccessHandlerFn,
                aFailureHandlerFn, aContext);
        req.onerror = Webvisum.Common.objMethod(this, '_handleConnectionError');
        this.logger.log('sending ' + aMethod + " " + aUrl + " data=" + aData);
        req.open(aMethod, aUrl, true);
        for (h in aHeaders) {
            // Webvisum.Common.log('header: '+h+"="+aHeaders[h]);
            req.setRequestHeader(h.toString(), aHeaders[h].toString());
        }
        req.send(aData);
    },
    _generateHandler : function(aSuccessHandlerFn, aFailureHandlerFn, aContext) {
        var self = this;
        return (function(e) {
            const r = e.target;
            self.logger.log('RESPONSE:'+r);
            self.state.setState('connect', false);
            if (r.readyState == 4 && r.status == 200) {
                self.logger.log('response content: \n' + r.responseText);
                var responseXml = r.responseXML;
                //self.last=resp;
                if (responseXml) {
                    var response = new Webvisum.Response(responseXml);
                    if (!response.isOk()) {
                        if (aFailureHandlerFn) aFailureHandlerFn(response, aContext);
                        else self._errorAction(response.getErrorText());
                    } else aSuccessHandlerFn(response, aContext);
                } 
                else self._errorAction("Missing XML data in the response.")
            } else
                self._errorAction('Unexpected response status:' + r.status
                        + '  ' + r.statusText);
        })
    },
    _handleConnectionError : function(e) {
        this.logger.log('CONNECTION ERROR:'+ e.target);
        this.state.setState('connect', false);
        var msg = "Server connection error";
        this._errorAction(msg);
    },
    _errorAction : function(aMsg) {
        if (aMsg) {
            this.state.addError(aMsg);
            this.logger.log('WARNING:' + aMsg);
        }
    }
}

Webvisum.Request = {
    API_PATH : 'rest/',
    http_method : 'GET',
    method : '', // could be a part of 'params'
    data : null, //http body
    headers : {},
    params : {},
    responseSuccessFn : null,
    responseFailureFn : null,
    key : null, //request primary key, used for request buffering [String],
    frozen : false,
    initFrom : function(aRequest) {
        return this.init(aRequest.host, aRequest.port, aRequest.username, aRequest.pwd)
    },
    init : function(aHost, aPort, aUsername, aPwd) {
        this.host = aHost;
        this.port = aPort;
        this.username = aUsername;
        this.pwd = aPwd;
        return this;
    },
    addParams : function(aHash) {
        for (var p in aHash)
            this.params[p] = aHash[p];
        return this;
    },
    setResponseHandler : function(anObject, aSuccesFnName, aFailureFnName) {
        if (aSuccesFnName)
            this.responseSuccessFn = Webvisum.Common.objMethod(anObject, aSuccesFnName);
        if (aFailureFnName)
            this.responseFailureFn = Webvisum.Common.objMethod(anObject, aFailureFnName);
    },
    setKey : function(aKeyString) {
        if (!this.fozen) {
            this.key = aKeyString;
        }
        return this
    },
    toString : function() {
        return this.toUrl() + ' (' + this.http_method + ') ';
    },
    toBaseUrl : function() {
        return this.host.match(/^http/) ? this.host : 'http://' + this.host
                + ':' + this.port + '/' + this.API_PATH;
    },
    toUrl : function() {
        const baseUrl = this.toBaseUrl();
        this.params['method'] = this.method;
        var querystring = '?';
        var pairs = new Array();
        for (p in this.params)
            pairs.push(escape(p) + '=' + escape(this.params[p]));
        return baseUrl + querystring + pairs.join('&');
    }
}

Webvisum.UploadRequest = Webvisum.beget(Webvisum.Request);
Webvisum.UploadRequest.include({
    http_method : 'POST',
    data : null,
    headers : {},
    setResourceUri : function(aResourceUri, aMimeType, aResourceName) {
        // create a stream from url given resource
        
        var io = Components.classes["@mozilla.org/network/io-service;1"]
            .getService(Components.interfaces.nsIIOService);
        var channel = io.newChannelFromURI(aResourceUri);
        var dataStream = channel.open();
        this.data = this.generateFileUploadData(dataStream, aMimeType, aResourceName);
    },    
    setCachedResource : function(aResourceUrl, aMimeType, aResourceName) {
        var streamFromCache = Webvisum.Monkey.fetchCacheData(aResourceUrl);
        Webvisum.Common.log('From Cache:'+streamFromCache);
        if (streamFromCache) {
            var dataStream = Components.classes["@mozilla.org/network/buffered-input-stream;1"]
                .createInstance(Components.interfaces.nsIBufferedInputStream);
            dataStream.init(streamFromCache,  streamFromCache.available());        
            this.data = this.generateFileUploadData(dataStream, aMimeType, aResourceName);
        }
        else alert('NOT IN CACHE');
    },    
    generateFileUploadData : function(aResourceInputStream, aMimeType, aResourceName) {
        const MULTI = "@mozilla.org/io/multiplex-input-stream;1";
        const STRINGIS = "@mozilla.org/io/string-input-stream;1";
        const nsIMultiplexInputStream = Components.interfaces.nsIMultiplexInputStream;
        const nsIStringInputStream = Components.interfaces.nsIStringInputStream;
        const BOUNDARY = "---------------------------21573107913958159401776129921";

        //mime header & parameters
        var header = new String();
        header += "\r\n";
        header += "--" + BOUNDARY + "\r\n";

        this.params['method'] = this.method;
        for (p in this.params) {
            header += "Content-Disposition: form-data; name=\"" + p
                    + "\"\r\n\r\n" + this.params[p] + "\r\n";
            header += "--" + BOUNDARY + "\r\n";
        }
        header += "Content-Disposition: form-data; name=\"file\"; filename=\""
                + aResourceName + "\"\r\n";
        header += "Content-Type: " + aMimeType + "\r\n";
        header += "\r\n";
        var header_stream = Components.classes[STRINGIS]
                .createInstance(nsIStringInputStream);
        header_stream.setData(header, header.length);

        // Ending MIME boundary
        var end_stream = Components.classes[STRINGIS]
                .createInstance(nsIStringInputStream);
        var bs = new String("\r\n--" + BOUNDARY + "--\r\n");
        end_stream.setData(bs, bs.length);
        // concatenate
        var mux_stream = Components.classes[MULTI].createInstance(nsIMultiplexInputStream);
        mux_stream.appendStream(header_stream);
        mux_stream.appendStream(aResourceInputStream);
        mux_stream.appendStream(end_stream);

        this.headers["Content-Type"] = "multipart/form-data; boundary="+ BOUNDARY;
        this.headers["Content-Length"] = mux_stream.available();
        return mux_stream
    },
    toUrl : function() {
        return this.toBaseUrl();
    }
});

Webvisum.Response = function(aXml) {
    XXX = aXml; //TODO: DEBUG CODE - remove
    this.corrupted = true;
    if ( aXml.documentElement.nodeName=='result'){
        this.root = Webvisum.Common.tag('result', aXml);    
        this.corrupted = (this.root == null);
        if (!this.corrupted) {
            this.status = this.root.getAttribute('status');
            this.method = this.root.getAttribute('method');
            this.message = this.root.getAttribute('message');
            this.corrupted = false;
        }
    }        
}

Webvisum.Response.prototype = {
    isOk : function() {
        return !this.corrupted && (this.status == 'ok');
    },
    hasServerMsg:function () {
        return this.message == '1'; 
    },
    getErrorCode:function () {
        if (this.corrupted) return -1;
        if (!this.isOk()) {
            var e = Webvisum.Common.tag('error', this.root);
            return 1*e.getAttribute('code');
        }
        return 0;
        
    },
    getErrorText : function() {
        if (this.corrupted)
            return "Corrupted server response";
        if (this.getErrorCode()!=0) {
            var e = Webvisum.Common.tag('error', this.root);
            return e.getAttribute('text') + ' (error:'+e.getAttribute('code')+')';
        }
        return '';
    },
    tagText : function(aTagName) {
        try {
            return Webvisum.Common.tag(aTagName, this.root).textContent;
        } catch (e) {
            Webvisum.Common.logErr("Response.tagText, tag " + this.root.nodeName
                    + " has not descendant tag " + aTagName);
            return null;
        }
    }
}