// REQUIRES: common.js, server.js, sound.js, state.js, urlfilter.js, protocol.js, ....



/*
 * Singleton object for entire application. Shared data for all opened browsers.
 * Accessible from WebvisumService XPCOM component.
 */
Webvisum.Global = {
    APP_VERSION : '0.7.7',
    state:null,
    db:null,
    prefs:null,
    server:null,
    visuals:null,
    clients:0, // refcount
    // called on startup of each client / browser
    // return instance of global webvisum
    onLocalUp:function (aClient) {
        this.log('Bootstrapping new Webvisum Client '+aClient);
        var globalWebvisum = Webvisum.getWebvisumService().global;
        if (!globalWebvisum) {
            this.startup();
            Webvisum.getWebvisumService().global = globalWebvisum = this;            
        }
        aClient.id = globalWebvisum.clients++;         
        this.log('Clients:'+globalWebvisum.clients);        
        return globalWebvisum
    },
    // called on closing of each client / browser
    //returns true if this was the last one
    onLocalDown:function (aClient) {
        this.log('Closing Webvisum Client '+aClient);
        this.clients--;
        var globalWebvisum = Webvisum.getWebvisumService().global;
        this.log('Clients:'+globalWebvisum.clients);
        if (globalWebvisum.clients<=0 && globalWebvisum){
            globalWebvisum.clients = 0;
            Webvisum.globalWebvisum.shutdown();
            getWebvisumService().global = null;
            return true;
        }
        return false;
    },
    startup:function () {
        this.log('Initializing GLOBAL Webvisum');
        //preferences
        var prefs = Components.classes["@mozilla.org/preferences-service;1"]
                .getService(Components.interfaces.nsIPrefService)
                .getBranch("extensions.webvisum.");
        this.prefs = prefs.QueryInterface(Components.interfaces.nsIPrefBranch2);
        Webvisum.Common.LOGGING = this.isDebug(); // global hack for logging
        //state
        this.state = Webvisum.beget(Webvisum.State);
        this.state.setState('disabled', !this.isActive());
        // urlFilter
        this.urlFilter = Webvisum.beget( Webvisum.UrlFilter );
        this.updateUrlFilter();
        //db
        this.db = Webvisum.beget( Webvisum.DB ).init();
        //server
        this.server = new Webvisum.Server(this.state, this);
        this.updateServer();
        //sound
        this.sound = Webvisum.beget( Webvisum.Sound ).init();
        this.initSounds();
        //visuals
        this.visuals = Webvisum.beget( Webvisum.Visuals );
        this.initVisuals();
        //captcha rules
        this.captchaRules = this.initCaptchaRules();
        //pref observer
        this.prefsObserver = Webvisum.beget( Webvisum.PreferencesGlobalObserver ).init(this);
        if (this.isActive() && this.isEULAaccepted()) this.activate();
    },
    shutdown:function () {
        this.log('Shutting down GLOBAL Webvisum');
        if (this.isActive()) this.deactivate();
        if (this.prefsObserver) this.prefsObserver.destroy();
        this.sound.shutdown();
        this.sound = null;
        this.visuals.removeAllDocuments();
        if (this.db) {
            this.db.destroy();
            this.db = null;
        }
        this.visuals = null;
        this.server = null;
        this.state = null;
        this.prefs = null;
    },
    activate:function () {
        this.log('Activating');
        this.server.enable();
    },
    deactivate:function(){
        this.log('Deactivating');
        // global, as the last instance should, logout
        if (this.state.isLogged()) {
            var protocol = new Webvisum.Protocol(this); //temporary protocol
            protocol.auth_logout_request();
        }
        
        if (this.server) this.server.disable();
    },
    initCaptchaRules:function () {
        this.log('loading cached captcha rules');
        const captchaString = this.prefs.getCharPref('captcha.rules');
        const rules = Webvisum.beget(Webvisum.CaptchaRules).fromString(captchaString);
        this.log("Loaded "+ rules.rules.length+" captcha rules from "+Webvisum.Common.toDate(rules.last_update));      
        return rules;
    },
    updateCaptchaRules:function (aCaptchaRules) {
        this.log('updating captcha rules');        
        const captchaString = aCaptchaRules.toXmlString(); 
        this.prefs.setCharPref('captcha.rules', captchaString);
        this.captchaRules = aCaptchaRules;
        return captchaString;
    },
    initVisuals:function () {
        var styleMap = {}
        for (var styleId in this.visuals.styles){
            styleMap[styleId] = this.prefs.getBoolPref( 'visuals.'+styleId );
        }
        this.visuals.initStyles( styleMap );
    },
    initSounds:function (optMuteOnly) {
        for (var soundId in this.sound.events){
            if (!optMuteOnly) { 
                var urlstring = this.prefs.getCharPref( 'sound.'+soundId );
                this.sound.setSound(soundId, urlstring);
            }
            this.sound.mute (soundId, !this.prefs.getBoolPref( 'sound_on.'+soundId ) );
        }
        if (!this.prefs.getBoolPref( 'sound_on')) {
            this.sound.muteAll();
        }
    },
    updateSound:function (aName) {
        if (this.sound){
            var urlstring = this.prefs.getCharPref( 'sound.'+aName );
            this.sound.setSound(aName, urlstring);
        }
    },
    updateServer:function () {
        if (this.server){
            const host = this.prefs.getCharPref( 'server.host' );
            const port = this.prefs.getCharPref( 'server.port' );
            const usr = this.prefs.getCharPref( 'server.username' );
            const pwd = Webvisum.Monkey.loadPassword(true);
            this.log("Updating server settings: "+host+', '+port+', '+usr+', '+pwd);
            this.server.setServer(host, port, usr, pwd);
        }
    },
    updateUrlFilter:function () {
        const whitelist = this.prefs.getCharPref( 'trigger.whitelist' );
        const blacklist = this.prefs.getCharPref( 'trigger.blacklist' );
        
        this.log("Updating url filter settings.");
        this.urlFilter.setRules(whitelist, blacklist);
    },        
    initServer:function () {
    },
    log:function (aMsg) {
        Webvisum.Common.log("[wsg]" + aMsg);
    },    
    get language () {
        var lang = this.prefs.getCharPref( 'general.language' );
        if (!lang || lang.length==0){
            lang = Webvisum.Common.getLocale();
        }
        return lang;
    },
    get global() {
        return this
    },

    isActive:function () {
        return this.prefs.getBoolPref( 'active' );  
    },
    isEULAaccepted:function () {
        //NOTE: temporarily disabled since version 0.7.3
        return true;
        var EULAVersion = this.prefs.getIntPref("EULA.version");
        return this.prefs.getBoolPref("EULA." + EULAVersion + ".accepted");
    },    
    isDebug:function () {
        return this.prefs.getBoolPref( 'debug' );  
    }
}


/*
 * One instance per each browser. 
 */
Webvisum.Local = function (){
    this.id = -1; //unregistered
}

Webvisum.Local.prototype = {
    startup:function () {
        Webvisum.Common.LOGGING = this.global.isDebug();
        this.log('Initializing LOCAL Webvisum');       
        //local preferences 
//        this.prefs = this.global.prefs;
        this.prefsObserver = Webvisum.beget( Webvisum.PreferencesLocalObserver ).init(this);
        //state        
        this.stateObserver = new Webvisum.StateLocalObserver(this);
        this.stateObserver.init();
        //GUI update
        this.updateButtonState();
        this.updateMenuState();
        this.updateMenuPrefs();
        this.toggleAutorefresh(true);
        this.toggleDynamicSelects(true);
        //temporary page filter
        this.temporaryFiler = Webvisum.beget(Webvisum.TemporaryFilter).init(this);
        //context menu/operations handler
        this.context = Webvisum.beget(Webvisum.ContextManager).init(this);
        //eula
        Webvisum.Common.fork( Webvisum.Common.objMethod(this, 'ensureEula'), 200); 
        
        //activation
        if (this.global.isActive() && this.global.isEULAaccepted()) this.activate();
        //see getters for aliases to global
    },
    shutdown:function () {
        this.log('Shutting down LOCAL Webvisum');    
        if (this.global.isActive()) this.deactivate();
        //context manager
        if (this.context){
            this.context.destroy();
            this.context = null;            
        }
        if (this.temporaryFiler) this.temporaryFiler.destroy();
        // state
        if (this.stateObserver) {
            this.stateObserver.destroy();
            this.stateObserver = null;
        }
        //preferences
        if (this.prefsObserver) this.prefsObserver.destroy();
    },
    activate:function () {
        this.log('Activating Webvisum');
        this.protocol = new Webvisum.Protocol(this);
        this.domListener = new Webvisum.DomListener(this).init();
        this.clickListener = new Webvisum.ClickListener(this).init();
        // key shortcuts
        this.keyShortcuts = Webvisum.beget( Webvisum.Keys ).init(this);
        this.initKeys();
        // make sure that the user is logged in if autologin
        // but only do this for the _first_ window, we don't want them being logged in everytime they open a new window
        var globalWebvisum = Webvisum.getWebvisumService().global;
        if (!globalWebvisum.autoLogginChecked){
          globalWebvisum.autoLogginChecked = true;
          Webvisum.Common.fork( Webvisum.Common.objMethod(this, 'ensureLogged'), 400);               
        }
    },
    
    deactivate:function(){
        this.log('Deactivating Webvisum');
        if (this.keyShortcuts) {
            this.keyShortcuts.destroy();
            this.keyShortcuts = null;
        }
        this.protocol = null;
        if (this.clickListener) {
            this.clickListener.destroy();
            this.clickListener = null;
        }
        if (this.domListener) {
            this.domListener.destroy();
            this.domListener = null;
        }
    },
    initKeys:function () {
        for (var keyEvent in this.keyShortcuts.keyBindings){
            var keyDef = this.prefs.getCharPref( 'keys.'+keyEvent );
            this.keyShortcuts.setKeyBinding(keyEvent, keyDef);
        }
    },
    // ---- ALIASES
    get global() {
        return Webvisum.getWebvisumService().global
    },
    get currentDocument() {
        const win = Webvisum.Common.unwrap(gBrowser.selectedBrowser.contentWindow);
        return win.wdocument
    },
    get currentBrowser() {
        return gBrowser.selectedBrowser;
    },
    get browser() {
        return gBrowser;
    },
    get db() {
        return this.global.db   
    },
    get server() {
        return this.global.server   
    },
    get prefs() {
        return this.global.prefs   
    },
    get sound() {
        return this.global.sound   
    },
    get language() {
        return this.global.language   
    },
    getWDocument:function (aTargetNode) {
        if (!aTargetNode) return true;
        var win = aTargetNode.defaultView || aTargetNode.ownerDocument.defaultView;
        win = Webvisum.Common.unwrap(win);
        return ('wdocument' in win) ? win.wdocument : null;
    },
    // translation
    m:function (aString, optArgArray) {
        if (!this.stringBundle) this.stringsBundle = document.getElementById("webvisum-strings");
        return optArgArray ?
            this.stringsBundle.getFormattedString(aString, optArgArray):
            this.stringsBundle.getString(aString);
    },
    toggleActive:function () {
        //check EULA acceptance, do not activate unless accepted
        var value = this.prefs.getBoolPref('active');
        if (!value && !this.ensureEula()) return;
        
        this.prefs.setBoolPref('active', !value);
    },
    log:function (aMsg) {
        Webvisum.Common.log("[wsl:"+this.id+"]" + aMsg);
    },
    logErr:function(aMsg, aDetails){
        Webvisum.Common.logErr("[wsl:"+this.id+"]" + aMsg, aDetails);
    },
    logWrn:function(aMsg, aDetails){
        Webvisum.Common.logWrn("[wsl:"+this.id+"]" + aMsg, aDetails);
    },
    toString:function () {
        return "Local Webvisum "+this.id;        
    },
    isAutologin:function () {
        return this.prefs.getBoolPref( 'server.autologin' );        
    },
    isAutotrigger:function () {
        return this.prefs.getBoolPref( 'general.autotrigger' );        
    },
    isFixDynamicSelects:function () {
        return this.prefs.getBoolPref( 'general.fix_dynamic_selects' );        
    },
    isVerbose:function () {
        return this.prefs.getBoolPref( 'general.verbose' );        
    },
    isModalNotifications:function () {
        return this.prefs.getBoolPref( 'general.modalnotifications' );        
    },
    // ------------------ ACTIONS -----------------
    login:function(optDelayMs) {
        var failureHandler = Webvisum.Common.objMethod(this, 'openLogin');
        this.protocol.auth_login_request(optDelayMs, failureHandler);
    },
    logout:function() {
        this.protocol.auth_logout_request();
    },
    fetch_messages:function(optDelayMs) {
        this.protocol.messages_read(optDelayMs);
    },
    cache_object:function (aWebvisumObject) {
        if (aWebvisumObject.isDirtyCache) {
            const domain = Webvisum.Monkey.extractDomainFromUrl(aWebvisumObject.page_url);
            this.log("Updating cache for '"+domain+ "' with "+ aWebvisumObject);
            this.db.updateObject(domain, aWebvisumObject.type, aWebvisumObject.id, aWebvisumObject.label, aWebvisumObject.language);
            aWebvisumObject.isDirtyCache = false;
        }
    },
    send_object:function (aWebvisumObject) {
        if (aWebvisumObject.isDirtyServer)
            this.protocol.objects_label_request(aWebvisumObject);
    },
    //manual update in all frames of this browser window
    manualDocumentHandling:function () {
        const rootDoc = this.currentBrowser.contentDocument;
        const allDocuments = Webvisum.Monkey.collectDocuments(rootDoc);
        for (var i = 0; i < allDocuments.length; i++) {
            var wdoc = this.getWDocument( allDocuments[i] );
            this.handleDocument( wdoc );
        }
    },
    handleDocument:function (optWebvisumDocument) {
        try{
            const wdoc = optWebvisumDocument ? optWebvisumDocument : this.currentDocument;
            this.log('Handling update of document '+wdoc);
            // is the page blocked?
            if (!this.global.urlFilter.isAllowed(wdoc.page_url)) {
                Webvisum.Common.log('Not invoking update on '+wdoc.page_url+' - BLOCKED');
                wdoc.isBlocked = true;
                return;
            }
            // is the document temporarily blocked?
            if (this.temporaryFiler.isBlocked(wdoc.window)) {
                Webvisum.Common.log('Not invoking update on '+wdoc.page_url+' - TEMPORARILY BLOCKED');
                wdoc.isBlocked = true;
                return;
            }
            //just a sound
            this.sound.playDefault();
            
            
            // install visuals
            this.global.visuals.addDocument(wdoc.document);
            
            //fixing dynamic selects
            if (this.isFixDynamicSelects()) this.fixDynamicSelects(wdoc.document);
            
            // start page annotation
            this.annotateDocument(wdoc);
            wdoc.isHandled = true;
        }
        catch(e) {
            this.logErr("Document handling error", e);
        }
    },
    annotateDocument:function (optWebvisumDocument) {
        const wdoc = optWebvisumDocument ? optWebvisumDocument : this.currentDocument;
        new Webvisum.WorkflowDBUpdate(this, wdoc).start();
    },
    blockCurrentDomain:function () {
        const docUrl = this.currentBrowser.contentDocument.location;
        if (!docUrl) return;
        const docDomain = Webvisum.Monkey.extractDomainFromUrl( docUrl );
        const pattern = "^(https?\:\/\/)?[\w\.]*"+docDomain;         
        var blacklist = this.prefs.getCharPref( 'trigger.blacklist' );
        const lines = blacklist.split("\n");
        for (var i = 0; i < lines.length; i++) {
            var line = Webvisum.Common.trim(lines[i]);
            if (line==pattern) return;
        }
        blacklist = pattern+"\n"+blacklist;
        this.prefs.setCharPref( 'trigger.blacklist', blacklist );
        
        this.notifyInfo( this.m("dialog.webvisum.blockdomain", [docDomain]) );
    },
    fixDynamicSelects:function (aDocument) {
        const selects = Webvisum.Monkey.findSensitiveSelects(aDocument);
        this.log('Fixing '+selects.length+' dynamic selects');
        const label = this.m('dialog.webvisum.fix_dynamic_selects');
        for (var selIdx = 0; selIdx < selects.length; selIdx++) {
            Webvisum.Monkey.fixSensitiveSelect(selects[selIdx], label);
        }
    },
    showOriginalPage:function () {            
        this.temporaryFiler.reload( this.currentBrowser );
    },
    ensureEula:function () {
        var displayEula = !this.global.isEULAaccepted();
        var accepted = true;
        if (displayEula) {
            accepted = this.openEULA();
            this.prefs.setBoolPref('active', accepted);
            this.sound.playDefault();
        }
        if(this.global.isActive()) {
            //dont show enabling msg if also logging in
            if(!this.isAutologin()) this.notifyInfo( this.m('dialog.webvisum.status.enabled'));
        }
        else if (this.isVerbose()) this.notifyInfo( this.m('dialog.webvisum.status.disabled'));
        return true
    },
    //asynchronous
    ensureLogged:function () {
      
        if (!this.global.state.isLogged()){
            if (this.isAutologin()) {
                this.log('Trying to autologin now.');
                return this.openLogin(false);
            }
            else {
                this.log("Not logged (no  autologin)");
                return false;//throw "Not logged (no  autologin)";
            }
        }
        return true;
    },
    
    //legacy function
    retrieveOldPassword: function(){
      var username = this.prefs.getCharPref( 'server.username' );
      var oldPassword = this.prefs.getCharPref( 'server.pwd' );
      if (oldPassword && username){
        //store the password in the password manager
        if (Webvisum.Monkey.savePassword(oldPassword)){
          //and remove from about:config 
          this.prefs.clearUserPref( 'server.pwd' );
        }
        return oldPassword;
      }
    },

    openLogin:function (isRetry) {

        if (!Webvisum.Monkey.enterMasterPassword()){
          //user refuses to enter their master password, no point in continuing as we can't save the password
          return false;
        }
      
        var cred = {
            username: this.prefs.getCharPref( 'server.username' ),
            password: Webvisum.Monkey.loadPassword(),
            checkbox: this.prefs.getBoolPref( 'server.autologin' )
        }
        
        
        //store old passwords in the login manager instead of about:config
        if (this.prefs.getPrefType('server.pwd')){
          cred.password = this.retrieveOldPassword();
        }
        
        if (cred.username && cred.password){
          Webvisum.Global.updateServer();
        }
        
        var retryLoop = isRetry;
        while ((cred.username.length==0) || (cred.password.length==0) || retryLoop){
            if (isRetry) this.log('Bad username or password. Prompting user.');
            else this.log('No usr/pwd. Prompting user.');
            
            var msg = isRetry ? this.m('dialog.webvisum.login.nonauthorized') : this.m('dialog.webvisum.login.nocredentials');
            cred = Webvisum.Common.promptUsernamePassword(this.m("dialog.webvisum.login.title"),
                msg, this.m("dialog.webvisum.login.remember"), cred);
            this.log('cred given');
            if (cred) {
              //store credentials
              this.prefs.setCharPref('server.username', cred.username);
              
              //save the password in the password manager
              Webvisum.Monkey.savePassword(cred.password);
              
              this.prefs.setBoolPref('server.autologin', cred.checkbox);
              
              //update the server credentials, bugger this doesn't work :(
              Webvisum.Global.updateServer();
              retryLoop = false;
            }
            else {
                this.log("Username prompt canceled. Logging skipped");
                return false;//throw "Not logged (no  autologin)";                      
            }
        }
        //login
        this.login();
        return true;
    },
    executeCommand:function (aCommand) {
        const map = {
            menu : "cmd_webvisum_menu",
            help : "cmd_webvisum_showhelp",
            label : "cmd_webvisum_handle_automatic",
            lastlink : "cmd_webvisum_handle_lastlink", 
            labelpage : "cmd_webvisum_handle_page",
            ocr : "cmd_webvisum_handle_ocr",
            captcha : "cmd_webvisum_handle_captcha",
            underline : "cmd_webvisum_visuals.underline",
            highlight : "cmd_webvisum_visuals.highlight",
            hideimages :  "cmd_webvisum_visuals.hide_images",
            visuals : "cmd_webvisum_visuals",
            sounds : "cmd_webvisum_sound_on",
            pageupdate : "cmd_webvisum_updatedocument",
            pageinfo : "cmd_webvisum_tools_pageinfo",
            status : "cmd_webvisum_status",
            activity : "cmd_webvisum_toggleactive"
        }
        if (aCommand in map) {
            const cmdId = map[aCommand];
            Webvisum.Common.idd(cmdId).doCommand();
        }
    },
    //---------------  GUI HANDLING
    onStatusbarTooltip:function (aTooltip) {             
        const state = this.global.state;
        Webvisum.Common.idd('webvisum_status_tooltip_headline').value = "Webvisum - "+state.getGlobalState();
        
        if (state.getErrors().length>0){
            var d = Webvisum.Common.idd('webvisum_status_tooltip_errors');
            d.value = state.errorsAsString();
            d.hidden=false;            
        }
        else {
            var d = Webvisum.Common.idd('webvisum_status_tooltip_errors');
            //d.value="----";
            d.hidden=true;            
        }
        return true;
    },
    togglePref:function(aCheckbox) {
        const checkbox = (aCheckbox.nodeName=='command') ? aCheckbox : Webvisum.Common.idd(aCheckbox.observes); 
        const propId = checkbox.getAttribute('propId');
        this.prefs.setBoolPref(propId, !this.prefs.getBoolPref(propId));
    },
    toggleAutorefresh:function (optJustUpdate) {
        try{
            const checkbox = Webvisum.Common.idd('cmd_webvisum_toggle_autorefresh');       
            var globalPrefs = Components.classes["@mozilla.org/preferences-service;1"]
                    .getService(Components.interfaces.nsIPrefService);
            var disabled = !globalPrefs.getBoolPref('accessibility.blockautorefresh');
            if (!optJustUpdate) disabled = !disabled; 
            globalPrefs.setBoolPref('accessibility.blockautorefresh', disabled);
            checkbox.setAttribute('checked', disabled);
        }
        catch (e){
            //swallow - does not work for FF2
        }
    },
    toggleDynamicSelects:function (optJustUpdate) {
        const checkbox = Webvisum.Common.idd('cmd_webvisum_toggle_dynamicselects');       
        var enabled = this.prefs.getBoolPref('general.fix_dynamic_selects');
        this.log('dynamic selects:'+enabled);
        if (!optJustUpdate) { 
            enabled = !enabled;
            this.prefs.setBoolPref('general.fix_dynamic_selects', enabled);
        }
        checkbox.setAttribute('checked', enabled);
    },
    onMenuShow:function () {
        const self = this;
        const dynAccel = function(anEvent){
            var keyDef = self.prefs.getCharPref('keys.'+anEvent);
            return Webvisum.Common.capitalize(keyDef).replace(/\ /g,'+')
        }
        
        Webvisum.Common.idd('webvisum_mi_help').setAttribute('acceltext', dynAccel('help') );
        Webvisum.Common.idd('webvisum_mi_toggleactive').setAttribute('acceltext', dynAccel('activity') );
        Webvisum.Common.idd('webvisum_mi_update').setAttribute('acceltext', dynAccel('pageupdate') );
        Webvisum.Common.idd('webvisum_mi_tools_pageinfo').setAttribute('acceltext', dynAccel('help') );
        Webvisum.Common.idd('webvisum_mi_sounds_mute_all').setAttribute('acceltext', dynAccel('sounds') );
        Webvisum.Common.idd('webvisum_mi_visual_any').setAttribute('acceltext', dynAccel('visuals') );
        Webvisum.Common.idd('webvisum_mi_visual_underline').setAttribute('acceltext', dynAccel('underline') );
        Webvisum.Common.idd('webvisum_mi_visual_highlight').setAttribute('acceltext', dynAccel('highlight') );
        Webvisum.Common.idd('webvisum_mi_visual_hideimages').setAttribute('acceltext', dynAccel('hideimages') );
        var debug = this.global.isDebug();
        Webvisum.Common.idd('webvisum_mi_update').hidden=!debug;
        Webvisum.Common.idd('webvisum_mi_cleanerrors').hidden=!debug; //disabled since 0.7.5
        Webvisum.Common.idd('webvisum_mi_toggleactive').hidden=!debug; //disabled since 0.7.5
        Webvisum.Common.idd('webvisum_menu_debug').hidden=!debug; 
        
    },
    onToolMenuPopup:function (aMenuElement) {
        var m=Webvisum.Common.idd('webvisum_status_menu'); 
        m.showPopup(aMenuElement.parentNode,-1,-1,'popup','topright','topleft'); 
        m.focus(); 
        
        m.listener = Webvisum.Common.objMethod(this, 'onToolMenuPopupKey');
        addEventListener("keydown", m.listener, true);
        return true;        
    },
    onToolMenuPopupHide:function (aMenuElement) {
        //this.log('hiding');
        var m=Webvisum.Common.idd('webvisum_status_menu');
        m.hidePopup();
        
        removeEventListener("keydown", m.listener, true);
        m.listener = null;
    },
    onToolMenuPopupKey:function (anEvent) {
      //because the menu is open independantly of the context menu, then it won't close when you hit 
      //the left arrow key (you'll need to hit escape). But we want it to act as if it _is_ connected
      //to the context menu. So we'll need to close the menu if a user hits the left arrow when the 
      //main menu is "focused"
      if (anEvent.keyCode ==  KeyEvent.DOM_VK_LEFT){
        var menu = document.getElementById('webvisum_status_menu');
        for(var i=0; i<menu.childNodes.length; i++){
          var child = menu.childNodes[i];
          if (child.nodeName.toLowerCase() == "menu" && child.menupopup && child.menupopup.state != "closed"){
            return true;
          }
        }
        //no submenu's are open, close the popup menu
        this.onToolMenuPopupHide();
        anEvent.stopPropagation();
        anEvent.preventDefault();
        return false;
      }
    },
    //KJD: use toolbarbuttons in Fx 4.x
    updateButtonState:function(){
      //use toolbarbuttons for Fx 4.x
      var appInfo = Components.classes["@mozilla.org/xre/app-info;1"].getService(Components.interfaces.nsIXULAppInfo);
      var version = parseInt(appInfo.version);
      if (version >= 4){
        //hide the statusbar 
        document.getElementById('webvisum_statusbar').hidden = true;
        //and enable the toolbarbutton if this is the first run
        if (!this.prefs.getBoolPref('toolbarbutton_installed')){
          //add the toolbarbutton to the current preference set
          this.addButtonToPalette('webvisum_toolbarbutton');
          this.prefs.setBoolPref('toolbarbutton_installed', true);
        }
      }
      else {
        //hide the toolbarbutton
        var btn = document.getElementById('webvisum_toolbarbutton');
        if (btn){
          btn.hidden = true;
        }
      }
    },
    //KJD: adds the toolbarbutton to the browsers toolbar palette
    //ref: https://developer.mozilla.org/en/Code_snippets/Toolbar#Adding_button_by_default
    addButtonToPalette:function(btnId){
      var afterId = "urlbar-container";    // ID of element to insert after
      var navBar  = document.getElementById("nav-bar");
      var curSet  = navBar.currentSet.split(",");

      if (curSet.indexOf(btnId) == -1){
        var pos = curSet.indexOf(afterId) + 1 || curSet.length;
        var set = curSet.slice(0, pos).concat(btnId).concat(curSet.slice(pos));

        navBar.setAttribute("currentset", set.join(","));
        navBar.currentSet = set.join(",");
        document.persist(navBar.id, "currentset");
        try {
          BrowserToolboxCustomizeDone(true);
        }
        catch(e){}
      }
    },
    updateMenuState:function(optChangedStateName) {
        const state = this.global.state;
        const status = state.getGlobalState();
        Webvisum.Common.idd('webvisum_statusbar', document).setAttribute('status', status);
        // logged
        const logged = !state.states['offline']; //(this.session_key!=null);
        Webvisum.Common.idd("cmd_webvisum_logout").setAttribute('disabled', !logged);
        Webvisum.Common.idd("cmd_webvisum_login").setAttribute('disabled', logged);
        Webvisum.Common.idd("cmd_webvisum_updatedocument").setAttribute('disabled', !logged || this.isAutotrigger() );
        
        // enabled/disabled
        const active = !state.states['disabled'];
        Webvisum.Common.idd('webvisum_mi_toggleactive').label = active ? this.m('menu.webvisum.cmd.disable') : this.m('menu.webvisum.cmd.enable');
        // show/hide relevant menu items
        const menuChilds = Webvisum.Common.idd('webvisum_status_menu').childNodes;
        for (var i = 0; i < menuChilds.length; i++) {
            if (menuChilds[i].getAttribute('class')!='permanent')
                menuChilds[i].hidden = !active;
        }
    },
    updateMenuPrefs:function (aPropName) {
        this.log('updating menu prefs');
        const self = this;
        const updater = function(aPropertyName, isInverse){
            var value = self.prefs.getBoolPref(aPropertyName);
            if (isInverse) value = !value;
            Webvisum.Common.idd('cmd_webvisum_'+aPropertyName).setAttribute('checked', value);
        }
        //visuals
        for (var styleId in this.global.visuals.styles) updater('visuals.'+styleId);            
        updater('visuals');
        //sounds
        for (var soundId in this.global.sound.events) updater('sound_on.'+soundId, true);            
        updater('sound_on', true);
        
        // FF bug workaround - last read property is passed as event to the next on change listner
        var prop = aPropName ? aPropName : 'visuals';
        this.prefs.getBoolPref(prop);
    },
    openPreferences:function() {
        window.openDialog("chrome://webvisum/content/preferences.xul", 
            "webvisum_preferences", "chrome,titlebar,toolbar,centerscreen,modal,resizable");
    },
    openHelp:function() {
        window.openDialog("chrome://webvisum/content/help.xul", 
            "webvisum_help", "chrome,titlebar,centerscreen,resizable");
    },
    //Opens modal EULA; returns true if accepted 
    openEULA:function() {
        const Cc = Components.classes, Ci = Components.interfaces;
        var ww2 = Cc["@mozilla.org/embedcomp/window-watcher;1"].getService(Ci.nsIWindowWatcher);
        var eulaWindow =  ww2.openWindow(null, "chrome://webvisum/content/EULA.xul",
            "_blank", "chrome,centerscreen,modal,resizable=yes", null);
        return this.global.isEULAaccepted();
    },
    notifyInfo:function (aMessage) {
        if (this.isModalNotifications()) Webvisum.Common.alertEx(aMessage, this.m('dialog.webvisum.title.info') );
        else this._notifyBox(aMessage, 0);
    },
    notifyWarning:function (aMessage) {
        if (this.isModalNotifications()) Webvisum.Common.alertEx(aMessage, this.m('dialog.webvisum.title.warn') );
        else this._notifyBox(aMessage, 5);
    },
    _notifyBox:function (aMessage, aType) {
        var box = this.browser.getNotificationBox();
        var priority = box.PRIORITY_WARNING_MEDIUM;
        switch(aType) {
            case 0: priority = box.PRIORITY_INFO_MEDIUM;
                break;
            case 5: priority = box.PRIORITY_WARNING_MEDIUM;
                break;
            case 10: priority = box.PRIORITY_CRITICAL_MEDIUM;
                break;
        }       
        box.removeTransientNotifications();
        box.appendNotification(aMessage, 'webvisum', null, priority, []);
    },    
    /*
     * Module getters
     */
    get tests() {
        return Webvisum.Tests.init(this)   
    }
}

/* 
 * DOM loading listener class
 * Each local webvisum has one instance.
 */
Webvisum.DomListener = function(aLocalWebvisum){
    this.webvisum=aLocalWebvisum
}
Webvisum.DomListener.prototype = {
    listener:null,
    init:function() {
        this.webvisum.log("initializing DOM Listener");
        var e = document.getElementById("appcontent");
        if(e) {
            this.listener = Webvisum.Common.objMethod(this, 'onDomLoaded');
            addEventListener("DOMContentLoaded", this.listener, false);
            return this;
        }
        return null;
    }, 
    destroy:function() {
        this.webvisum.log("removing DOM Listener");
        var e = document.getElementById("appcontent");
        if(e && this.listener) { 
            removeEventListener("DOMContentLoaded", this.listener, false);
            this.listener = null;
        }
    }, 
    /* Fired every time some DOM finises loading */
    onDomLoaded:function(event) {
        if (!this.listener) return;
//        const doc = Webvisum.Common.unwrap(event.target);
        const doc = event.target;
        this.webvisum.log('DOM loaded:'+doc.location);
        const wdoc = new Webvisum.Document(this.webvisum, doc);
        this.webvisum.log('Created '+wdoc);
        
        //check credentials excctraction
        if (this.checkCredentials(wdoc)) return;
        
        // decoroate the window
        Webvisum.Common.unwrap(wdoc.window).wdocument = wdoc;

        // install document unload handler (e.g. to remove from visuals)
        wdoc.window.addEventListener("unload", Webvisum.Common.objMethod(this,'onDocUnload'), false);

        if (!this.webvisum.isAutotrigger()) {
            Webvisum.Common.log('Not invoking update on '+doc.location+' - AUTOTRIGGER OFF');
            return;
        }        
        
        this.webvisum.handleDocument(wdoc);        
    },
    // Fired on window closed
    onDocUnload:function (event) {
//        const win = Webvisum.Common.unwrap(event.currentTarget);
        const win = event.currentTarget;
        const doc = win.document;
        this.webvisum.log('UNLOADING:'+doc.location+', wdoc:'+win.wdocument);
        
        if (this.webvisum.global){
            // remove from visuals
            this.webvisum.global.visuals.removeDocument(doc);
        }
    },
    checkCredentials:function (aWebvisumDocument) {
        if (aWebvisumDocument.window.name=='webvisum_registration') {
            var creds = Webvisum.Monkey.extractCredentials(aWebvisumDocument.document);
            if (creds) {
                if (this.webvisum.global.state.isLogged()) this.webvisum.logout();
                this.webvisum.prefs.setCharPref('server.username', creds.username);
                Webvisum.Monkey.savePassword(creds.password);
                //modal
                Webvisum.Common.alertEx( this.webvisum.m('dialog.webvisum.login.credsExtracted',[creds.username]), 
                    this.webvisum.m('dialog.webvisum.title.info') );
                aWebvisumDocument.window.close();                
            }
            return true
        }   
        return false
    }
}

Webvisum.ClickListener = function(aLocalWebvisum){
    this.webvisum=aLocalWebvisum
}
Webvisum.ClickListener.prototype = {
    listener:null,
    init:function() {
        this.webvisum.log("initializing Click Listener");
        var browser = this.webvisum.browser;
        if(browser) {
            this.listener = Webvisum.Common.objMethod(this, 'onClick');
            browser.addEventListener("click", this.listener, true);
            return this;
        }
        return null;
    }, 
    destroy:function() {
        this.webvisum.log("removing Click Listener");
        var browser = this.webvisum.browser;
        if(browser && this.listener) {
            browser.removeEventListener("click", this.listener, true);
            this.listener = null;
        }
    }, 
    onClick:function(anEvent) {
        if (!this.listener) return;
        //is left or middle click
        if (anEvent.button>1) return;
        const target = anEvent.originalTarget;
        const anchor =  Webvisum.Monkey.findOuterAnchor(target);
        this.webvisum.log('CLICK:'+ target + '  anchor '+anchor);
        if (anchor) this.webvisum.context.setLastAnchor(anchor);
    }
}


Webvisum.StateLocalObserver = function(aLocalWebvisum){
    this.webvisum=aLocalWebvisum
}
Webvisum.StateLocalObserver.prototype = {
    listener:null,
    init:function() {
        this.webvisum.log("initializing StateLocalListener");
        this.listener = Webvisum.Common.objMethod(this, 'onStateChange');
        this.webvisum.global.state.addListener(this.listener);
    }, 
    destroy:function() {
        this.webvisum.log("removing StateLocalListener");
        if (this.listener)
            this.webvisum.global.state.removeListener(this.listener);
        this.listener = null;
    }, 
    /* Fired on global state change, see state.js */
    onStateChange:function(aChangedStateName, aStateManager, optPayload) {
        if (!this.listener) return;
        if (/^msg.error/.test(aChangedStateName)) {
            this.webvisum.notifyWarning('Webvisum:'+optPayload);            
        }
        else if (/^msg.server/.test(aChangedStateName)) {
            this.webvisum.notifyInfo('Webvisum Server Message:'+optPayload);
        }
        else {
            this.webvisum.log('Global state changed: '+aChangedStateName);
            this.webvisum.updateMenuState(aChangedStateName);
            if (this.webvisum.isVerbose()){
                const casus = aChangedStateName+':'+aStateManager.states[aChangedStateName];
                var msg = null;
                switch (casus){
                    case 'disabled:true': msg = 'dialog.webvisum.status.disabled';break; 
                    case 'disabled:false': msg = 'dialog.webvisum.status.enabled';break;                    
                    case 'offline:true': msg = 'dialog.webvisum.status.loggedout';break; 
                    case 'offline:false': msg = 'dialog.webvisum.status.loggedin';break; 
                }
                if (msg) this.webvisum.notifyInfo( this.webvisum.m(msg));
            }
        }
    }    
}


/*
 * 
 */

//Abstract prototype
Webvisum.PreferencesObserver = {
    init:function(aWebvisum) {
        this.webvisum=aWebvisum;
        this.webvisum.log("initializing Preferences observer");
        var prefs = this.webvisum.prefs.addObserver("", this, false);
        return this;
    },
    destroy:function(){
        this.webvisum.log("removing Preferences Observer");
        this.webvisum.prefs.removeObserver("", this);
    },
    //subclass responsibility
    observe: function(subject, topic, data) {
        if (topic != "nsPref:changed") return;
        this.webvisum.log("Changed '"+data+"'");
    },
    fixFFbug:function (aPropName) {
        try{
            this.webvisum.prefs.getBoolPref(aPropName)
        }
        catch (e){            
            try {
                this.webvisum.prefs.getCharPref(aPropName)
            }
            catch(e){}
        }
    }
}

Webvisum.PreferencesLocalObserver = Webvisum.beget(Webvisum.PreferencesObserver);
Webvisum.PreferencesLocalObserver.include({
    observe: function(aSubject, aTopic, data) {
        if (aTopic != "nsPref:changed") return;
        this.webvisum.log("Changed(L) '"+data+"'");
        //= '"+this.webvisum.prefs.getCharPref(data)+"'");
        var names = data.split('.');
        
        switch(names[0]) {
            case "general": 
                this.webvisum.updateMenuState();
                break;
            case "visuals":
                this.handleVisuals(names[1], data);
                break;
            case "sound_on":
                this.handleSounds(names[1], data);
                break;
            case "keys":
                this.handleKeys(names[1], data);
                break;
            case "active":
                this.handleActivation();
                break;
            default:
        }
    },
    handleActivation:function () {
        if ( this.webvisum.global.isActive() ) this.webvisum.activate(); 
        else this.webvisum.deactivate();
        this.fixFFbug('active');
    },    
    handleKeys:function (aKeyEvent, aPropName) {
        const keyDef = this.webvisum.prefs.getCharPref( 'keys.'+aKeyEvent );
        this.webvisum.keyShortcuts.setKeyBinding(aKeyEvent, keyDef);
    },
    handleVisuals:function (aStyleName, aPropName) {
        this.webvisum.updateMenuPrefs(aPropName);
    },
    handleSounds:function (aStyleName, aPropName) {
        this.webvisum.updateMenuPrefs(aPropName);
    }    
})

Webvisum.PreferencesGlobalObserver = Webvisum.beget(Webvisum.PreferencesObserver);
Webvisum.PreferencesGlobalObserver.include({
    observe: function(aSubject, aTopic, aData) {
        if (aTopic != "nsPref:changed") return;
        this.webvisum.log("Changed (G)  '"+aData+"'");
        //= '"+this.webvisum.prefs.getCharPref(data)+"'");
        var names = aData.split('.');
        
        switch(names[0]) {
            case "sound":
                this.webvisum.updateSound(names[1]);
                break;
            case "server":
                this.webvisum.updateServer();
                break;
            case "trigger":
                this.webvisum.updateUrlFilter();
                break;
            case "visuals":
                this.handleVisuals(names[1], aData);
                break;
            case "sound_on":
                this.handleSounds(names[1], aData);
                break;
            case "active":
                this.handleActivation();
                break;
            default:
        }
    },
    handleActivation:function () {
        this.webvisum.state.setState('disabled', !this.webvisum.isActive());
        if ( this.webvisum.isActive() ) this.webvisum.activate(); 
        else this.webvisum.deactivate();
        this.fixFFbug('active');
    },    
    handleVisuals:function (aStyleId, aPropName) {
        const checked = this.webvisum.prefs.getBoolPref(aPropName);
        if (aPropName=='visuals'){
            //handle all
            if (checked) this.webvisum.initVisuals();
            else this.webvisum.visuals.disableAllStyles();            
        }
        else{
            if (checked) this.webvisum.visuals.enableStyle(aStyleId);
            else this.webvisum.visuals.disableStyle(aStyleId);
        }
    },    
    handleSounds:function (aSoundId, aPropName) {
        const checked = this.webvisum.prefs.getBoolPref(aPropName);
        if (aPropName=='sound_on'){
            //handle all
            if (checked) this.webvisum.initSounds(true);
            else this.webvisum.sound.muteAll();            
        }
        else{
            this.webvisum.sound.mute(aSoundId, !checked);
        }
    }    
})

    /* navigation Listener*/
//     navigationListener: {
//         init:function(aWebvisum) {
//            this.webvisum=aWebvisum
//            Webvisum.Common.log("initializing Navigation listener");
//            gBrowser.addProgressListener( this, Components.interfaces.nsIWebProgress.NOTIFY_STATE_DOCUMENT);
//        }, 
//        destroy:function(){
//            Webvisum.Common.log("removing Navigation listener");
//            gBrowser.removeProgressListener( this);
//        }, 
//        onStateChange: function(a, b, flags, d) {
//            if(flags & Components.interfaces.nsIWebProgressListener.STATE_START) {
//                Webvisum.Common.log('navigation loaded');
//            }
//        },
//        onLocationChange: function(a, b, c) {},
//        onProgressChange: function(a, b, c, d, e, f) {},
//        onSecurityChange: function(a, b, c) {},
//        onStatusChange: function(a, b, c, d) {},
//        onLinkIconAvailable: function(a, b) {},
//        QueryInterface: function(id) {
//          if(id.equals(Components.interfaces.nsIWebProgressListener) ||
//             id.equals(Components.interfaces.nsISupportsWeakReference) ||
//             id.equals(Components.interfaces.nsISupports))
//            return this;
//          throw Components.results.NS_NOINTERFACE;
//        }
//    },
//    /* download Observer*/
//    downloadObserver: {
//        init:function(aWebvisum) {
//            this.webvisum=aWebvisum
//            Webvisum.Common.log("initializing Download listener");
//            Components.classes["@mozilla.org/observer-service;1"]
//                .getService(Components.interfaces.nsIObserverService)
//                .addObserver(this, "dl-done", false);
//        }, 
//        destroy:function(){
//            Webvisum.Common.log("removing Download listener");
//            Components.classes["@mozilla.org/observer-service;1"]
//                .getService(Components.interfaces.nsIObserverService)
//                .removeObserver(this, "dl-done", false);
//        },         
//        observe: function(subject, topic, data) {
//            if(topic == "dl-done") {
//                Webvisum.Common.log('download finished loaded:'+subject + ", data" + data);
//            }
//        }
//    },


Webvisum.TemporaryFilter = {    
    init:function (aWebvisum) {
        this.webvisum = aWebvisum;
        this.protectedWindows = [];
        return this;
    },
    destroy:function () {
        this.webvisum = null;
        this.protectedWindows = [];        
    },
    reload:function (aBrowser) {
        var topWin = Webvisum.Common.unwrap(aBrowser.contentWindow);
        this.webvisum.log('Showing original version of the current page '+topWin.location);
        const self = this;
        const sweeper = function(){
            //remove top window
            self.webvisum.log('Sweeping temprorarly protected window: '+topWin.location)
            for (var i = 0; i < self.protectedWindows.length; i++) {
                if (self.protectedWindows[i]==topWin) {
                    self.protectedWindows.splice(i,1);
                    return;
                }
            }
        }
        self.protectedWindows.push(topWin);
        aBrowser.reload();
        Webvisum.Common.fork( sweeper, 10*1000);        
    },
    isBlocked:function (aWindow) {
        var topWin = Webvisum.Common.unwrap(aWindow.top); //handle also frames
        for (var i = 0; i < this.protectedWindows.length; i++) {
            if (this.protectedWindows[i]==topWin) return true;
        }
        return false;
    }
}


/**
 */
Webvisum.WorkflowDBUpdate = function(aWebvisum, aWebvisumDocument) {
    this.webvisum = aWebvisum;
    this.wdocument = aWebvisumDocument;
    if (!(aWebvisum && aWebvisumDocument)) throw "No documents"
}

Webvisum.WorkflowDBUpdate.prototype = {
    SERVER_BUFFER_DELAY: 500,
    TOO_OLD_THRESHOLD: 30*24*60*60, //in seconds
    //SPLIT + TERMINAL NODE
    start:function () {
        // query server for updated in paraller
        Webvisum.Common.fork( Webvisum.Common.objMethod(this, 'queryServer') );
        //update page from cached data
        this.queryDB();
    },
    // TERMINAL
    end:function () {
        //clean resources/mem if any
    },
    queryServer:function () {
        // get domain's last update
        var lastUpdate = this.webvisum.db.getLastUpdate(this.wdocument.domain);
        //fetch complete list if the cache is older than TOO_OLD_THRESHOLD 
        this.lastUpdate = ((lastUpdate + this.TOO_OLD_THRESHOLD)*1000 > 1*new Date() ) ? lastUpdate : 0;
        this.webvisum.log("Last update of:'"+this.wdocument.domain+"' was "+lastUpdate + " = "+Webvisum.Common.toDate(lastUpdate) +"  after thershold:" +this.lastUpdate);
        // fetch page update 
        this.webvisum.protocol.domains_getinfo_request(this.wdocument.page_url, this.wdocument.domain,
            Webvisum.Common.objMethod(this, 'onServerUpdate'), this.lastUpdate, this.SERVER_BUFFER_DELAY);
    },
    onServerUpdate:function(aWebvisumObjects, aDomain, aTimestamp, aPageUrl, aLanguage){
        //upload objects
        if (aDomain != this.wdocument.domain) {
            this.webvisum.logErr('Server returned other domain info ('+aDomain+') as requested ('+this.wdocument.domain+')');
            return;
        }
        //invalidate cache if fetching complete object list
        if (this.lastUpdate<=0) {
            this.webvisum.log('Invalidating cache...');
            this.webvisum.db.cleanCache(aDomain, aLanguage);
        }
        
        this.webvisum.log('Updating '+aWebvisumObjects.length+' objects');
        for (var i = 0; i < aWebvisumObjects.length; i++) {
            var obj = aWebvisumObjects[i];
            this.webvisum.db.updateObject(aDomain, obj.type, obj.id, obj.label, aLanguage);
        }  
        //update domain
        this.webvisum.log('DB objects updated. Updating DB domain timestamp to '+aTimestamp);
        this.webvisum.db.updateDomain(aDomain, aTimestamp);
        
        //apply objects
        this.applyObjects(aWebvisumObjects);
        this.wdocument.meta.annotationSrc  = 'server';
    },
    queryDB:function () {
        const pageUrl = this.wdocument.page_url;
        const domain =  this.wdocument.domain;
        const language = this.webvisum.language;
        const results = this.webvisum.db.getAllObjects(domain, language);
        this.webvisum.log("DB matches for '"+pageUrl+"'  domain:'"+domain+"' "+results.length+" db matches"); 
        if (results.length==0) {
            this.webvisum.log('No cached objects to update the page.');
            this.end();
        }
        var objects = [];
        for (var i = 0; i < results.length; i++) {
            var r = results[i];
            var o = Webvisum.Object.FactorFromArgs(r.type, r.objectid, r.label, language, pageUrl);
            objects.push(o);
        }
        this.applyObjects(objects);
        this.wdocument.meta.annotationSrc  = 'cache';
    },
    //OR-JOIN
    applyObjects:function (aWebvisumObjects) {
        //TODO: handle concurrency
        this.webvisum.log("Applying objects");
        for (var i = 0; i < aWebvisumObjects.length; i++) {
            var o = aWebvisumObjects[i];
            this.webvisum.log("Applying: "+o);
            try {
                o.apply(this.wdocument);
            } 
            catch(e) {
                this.webvisum.logErr("Cannot completely apply labling of "+o, e);
            }

        }
        this.end();
    }
};

/*
 * Inner Instance Module: Tests
 * getter: webvisum.tests 
 */
Webvisum.Tests = {
    init:function (aContainer) {
        this.webvisum = aContainer;
        return this;
    },
    //ref: https://developer.mozilla.org/en-US/Add-ons/Code_snippets/Tabbed_browser#Opening_a_URL_in_a_new_tab
    openWhiteboard: function(callback){
        var tab = gBrowser.addTab('chrome://webvisum/content/document-info.html');
        var browser = gBrowser.getBrowserForTab(tab);
        var onloadListener = function(){
            browser.removeEventListener("load", onloadListener, true);
            callback(browser.contentDocument);
        }
        browser.addEventListener("load", onloadListener, true);
        gBrowser.selectedTab = tab;
    },
    setDomainTimestamp:function (aTimestamp) {
        const wdoc = this.webvisum.currentDocument;
        const domain = wdoc.domain;
        var t = aTimestamp;
        if (!t) {
            t = prompt("Timestamp for actual domain '"+domain+"':", Math.floor(new Date()/1000));
        }
        this.webvisum.db.updateDomain(domain, t);
    },
    domainTimestamp:function () {
        const wdoc = this.webvisum.currentDocument;
        const domain = wdoc.domain;
        const lastUpdate = this.webvisum.db.getLastUpdate(domain);
        alert('Last update of:'+domain+" is "+lastUpdate+ "\n"+Webvisum.Common.toDate(lastUpdate));
    },
    cleanCache:function () {       
        const wdoc = this.webvisum.currentDocument;
        const domain = wdoc.domain;
        const language = this.webvisum.language;
        Webvisum.Common.log('Cleaning cache for domain '+domain+', lang='+language);
        this.webvisum.db.cleanCache(domain, language);
    },
    dbDomainInfo:function () {
        const wdoc = this.webvisum.currentDocument;
        const domain = wdoc.domain;
        const pageUrl = wdoc.page_url;
        const language = this.webvisum.language;
        const results = this.webvisum.db.getAllObjects(domain, language);

        this.openWhiteboard(function(doc){
            doc.title = "DBG:DBDomain";

            Webvisum.Monkey.html([
                {'pre': {'#text': "DEBUG: DB domaininfo for: '"+pageUrl+"' domain:'"+domain+"', lang:"+language+" -> "+results.length+" db matches"}}
            ], doc.getElementById('content'));
            
            var objects = [];
            for (var i = 0; i < results.length; i++) {
                var r = results[i];
                //Webvisum.Common.printObj(r);
                var o = Webvisum.Object.FactorFromArgs(r.type, r.objectid, r.label, language, pageUrl);
                objects.push(o);

                Webvisum.Monkey.html([
                    {'li': {'#text': i +". "+ o}}
                ], doc.getElementById('content'));
            }
        });
    },

    serverDomainInfo:function () {
        const pageUrl = this.webvisum.currentDocument.page_url;
        var self=this;
        const handler = function(aWebvisumObjects, aDomain, aTimestamp, aPageUrl, aLanguage){
            self.openWhiteboard(function(doc){
                doc.title = "DBG:ServerDomain";
                Webvisum.Monkey.html([
                    {'label': {'#text': 'DEBUG: Server domaininfo response for '+ pageUrl}}      
                ], doc.getElementById('subheading'));

                Webvisum.Monkey.html([
                    {'pre': {'#text': "domain:'"+aDomain+"'\npageUrl:"+aPageUrl+"\nlang:"+aLanguage+"\nupdate:"+aTimestamp+"="+Webvisum.Common.toDate(aTimestamp)+"\nobjs:"+aWebvisumObjects.length}},
                ], doc.getElementById('content'));

                for (var i = 0; i < aWebvisumObjects.length; i++) {
                    Webvisum.Monkey.html([
                        {'li': {'#text': i +'. '+ aWebvisumObjects[i]}},
                    ], doc.getElementById('content'));
                }
            });
        }

        //using null instead of domain to prevent frequency check 
        this.webvisum.protocol.domains_getinfo_request( pageUrl, null, handler, 0, 0 );
    },

    serverCaptchaRules:function () {
        var self=this;
        const handler = function(aCaptchaRules){
            self.openWhiteboard(function(doc){
                doc.title = "DBG:ServerCaptchaRules";
                Webvisum.Monkey.html([
                    {'label': {'#text': 'DEBUG: Server response for get captcha rules'}}      
                ], doc.getElementById('subheading'));

                Webvisum.Monkey.html([
                    {'div': {'#text': "update:"+ Webvisum.Common.toDate(aCaptchaRules.last_update) +" min_weight:"+ aCaptchaRules.min_weight +" rules:"+ aCaptchaRules.rules.length}},
                ], doc.getElementById('content'));
                
                for (var i = 0; i < aCaptchaRules.rules.length; i++) {
                    var rule = aCaptchaRules.rules[i];
                    Webvisum.Monkey.html([
                        {'pre': {'#text': "pattern='"+rule.pattern+"' weight="+rule.weight}},
                    ], doc.getElementById('content'));
                }
            });
        }        
        this.webvisum.protocol.captcha_getrules_request(0, handler, 0);
    },

    detectCaptcha:function () {
        var pageDoc = this.webvisum.currentDocument.document;
        var candidates = Webvisum.Monkey.findCaptchas(this.webvisum.global.captchaRules, pageDoc);
        
        this.openWhiteboard(function(doc){
            doc.title = "DBG: Captcha Hunter";
            Webvisum.Monkey.html([
                {'label': {'#text': 'DEBUG: Captcha detection on page (and subframes):'}},
                {'a': {'href': pageDoc.location, '#text': pageDoc.location}},
            ], doc.getElementById('subheading'));

            Webvisum.Monkey.html([
                {'p': {'#text': "Found "+ candidates.length +" candidates:"}}
            ], doc.getElementById('content'));


            //dump objects
            if (candidates.length){
                var rows = [];
                var count = 1;
                for (var i = 0; i < candidates.length; i++) {
                    var candidate = candidates[i];
                    var msg = candidates.toString();
                    var canvas = Webvisum.Monkey.image2canvas(candidate.target);
                    var imageUri = Webvisum.Monkey.canvas2dataurl(canvas);

                    var row = {
                        'tr': {'#children': [
                            {'td': {'#text': i+1}},
                            {'td': {
                                'class': i==0 ? "first-candidate" : null, 
                                '#text': msg
                            }},
                            {'td': {'#children': [
                                {'img': {
                                    'src': imageUri,
                                    'align': 'right' 
                                }}
                            ]}},
                        ]}
                    }
                    rows.push(row);
                }

                Webvisum.Monkey.html({'table': {'#children': rows}}, doc.getElementById('content'));                        
            }
        });
    },

    serverImageUpload:function () {
        var doc = this.webvisum.currentDocument.document;
        var img = doc.images[0];
        this.webvisum.log('first image from doc: '+img.src);
        var canvas = Webvisum.Monkey.image2canvas(img);
        var imageUri = Webvisum.Monkey.canvas2uri(canvas);
        var objectUrl = img.src;
        var mimeType = Webvisum.Monkey.guessMimeType(objectUrl);
        this.webvisum.log('Uploading image '+objectUrl);
        
        var rq = this.webvisum.server.newRequest('image.submit', {
            session_key : this.webvisum.global.state.session_key,
            page_url : objectUrl,
            object_id : objectUrl
        }, Webvisum.UploadRequest.initFrom(this.webvisum.server.genericRequest) );
        rq.setResourceUri(imageUri, mimeType, 'file');
        rq.setResponseHandler(this, 'serverResponse');

        this.webvisum.server.sendRequest(rq, null);
    },
    serverResponse:function (aResponse) {
        this.webvisum.log('.');
    },
    canvas:function () {
        if (gContextMenu && gContextMenu.onImage) {            
            var img = gContextMenu.target;
            this.webvisum.log('image form context:'+img.src);
        }
        else {
            var doc = this.webvisum.currentDocument.document;
            var img = doc.images[0];
            this.webvisum.log('first image from doc: '+img.src);
        }
        
        if (img) {
            this.webvisum.log("Drawing image "+img.src );
            var canvas = Webvisum.Monkey.image2canvas(img);
            Webvisum.Monkey.canvas2file(canvas, '/tmp/image'+1*new Date()+'.png');
        }
        else alert('no img found');
    },
    fixSensitiveSelects:function () {
        const rootDoc = this.webvisum.currentBrowser.contentDocument;
        const allDocuments = Webvisum.Monkey.collectDocuments(rootDoc);
        for (var i = 0; i < allDocuments.length; i++) {
            var doc = allDocuments[i];
            this.webvisum.fixDynamicSelects(doc);
        }
    },



    documentInfo: function(){
        const rootDoc = this.webvisum.currentBrowser.contentDocument;
        var self = this;

        self.openWhiteboard(function(doc){
            Webvisum.Monkey.html([
                {'label': {'#text': 'Root page Url:'}},
                {'a': {href: rootDoc.location, '#text': rootDoc.location}},                
            ], doc.getElementById('subheading'));

            Webvisum.Monkey.html("Last anchor :"+ self.webvisum.context.lastAnchor, doc.getElementById('last-anchor'));

            const allDocuments = Webvisum.Monkey.collectDocuments(rootDoc);
            for (var i = 0; i < allDocuments.length; i++) {
                var wdoc = this.webvisum.getWDocument( allDocuments[i] );

                Webvisum.Monkey.html([
                    {'hr': {}},
                    {'h4': {'#text': 'Document '+ (i + 1)}}
                ], doc.getElementById('content'));

                if (wdoc){
                    Webvisum.Monkey.html({'ul': {
                        'class': 'webvisum-doc-info',
                        '#children': [
                            {
                                'li': {'#children': [
                                    {'span': {'#text': 'Page url: '}}, 
                                    {'a': {'href' : wdoc.page_url, '#text': wdoc.page_url}}
                                ]}
                            },
                            {
                                'li': {'#children': [
                                    {'span': {'#text': 'Parent Window: '}}, 
                                    {'a': {'href' : (wdoc.window != wdoc.window.parent) ? wdoc.window.parent.location : null, '#text': (wdoc.window != wdoc.window.parent) ? wdoc.window.parent.location : 'No parent'}}
                                ]}
                            },
                            {
                                'li': {'#children': [
                                    {'span': {'#text': 'Domain: '}}, 
                                    {'code': {'#text': wdoc.domain}},
                                ]}
                            },
                            {
                                'li': {'#children': [
                                    {'span': {'#text': 'Was annotated/handled: '}}, 
                                    {'strong': {'#text': wdoc.isHandled}},
                                ]}
                            },
                            {
                                'li': {'#children': [
                                    {'span': {'#text': 'Annotation source: '}}, 
                                    {'strong': {'#text': wdoc.meta.annotationSrc}},
                                ]}
                            },
                            {
                                'li': {'#children': [
                                    {'span': {'#text': 'Was blocked: '}}, 
                                    {'strong': {'#text': wdoc.isBlocked}},
                                ]}
                            },
                        ]
                    }}, doc.getElementById('content'));

                    //dump objects
                    if (wdoc.objects){
                        var rows = [];
                        var count = 1;
                        for ( var objId in wdoc.objects) {
                            var obj = wdoc.objects[objId];
                            var row = {
                                'tr': {'#children': [
                                    {'td': {'#text': count++}},
                                    {'td': {
                                        'class': obj.target ? "found" : "notfound", 
                                        '#children': [
                                            {'code': {'#text': objId}}
                                        ]
                                    }},
                                    {'td': {'#text': obj.targetCount}},
                                    {'td': {'#text': obj, 'class': 'data'}}    
                                ]}
                            }
                            rows.push(row);
                        }

                        Webvisum.Monkey.html([
                            {'table': {
                                '#children': rows
                            }},
                        ], doc.getElementById('content'));                        
                    }
                }
                else {
                    Webvisum.Monkey.html({'div': {
                        'class': 'notfound',
                        '#text': 'No wdoc. Url: '+ allDocuments[i].location
                    }}, doc.getElementById('content'));
                }
            }
        });
    }
};
